import unittest

from lxml import etree as et

from eml.resources import EMLDistribution
from eml.resources.distribution import EMLOnline, EMLOffline
from test_xml.test_xml import TestXML


class TestDistribution(TestXML):
    DEFAULT_TAGS = {
        "scope": "document",
        "function": "download",
    }

    def test_parse_none(self):
        self.assertIsNone(EMLDistribution.parse(None, {}), "Distribution from nowhere")

    def test_no_argument(self):
        self.assertRaises(
            TypeError,
            EMLDistribution,
        )

    def test_multiple_distribution(self):
        self.assertRaises(
            TypeError,
            EMLDistribution,
            online=EMLOnline(),
            offline=EMLOffline("Tape, 3.5 inch Floppy Disk, hardcopy"),
        )

    def test_parse_online(self):
        xml_text = """
<distribution id="1" scope="document">
    <online>
        <url>http://data.org/getdata?id=98332</url>
    </online>
</distribution>
        """
        distribution = EMLDistribution.from_string(xml_text)
        self.assertEqual("1", distribution.id, "Error on parsing id")
        self.assertEqual(
            "http://data.org/getdata?id=98332",
            distribution.online.url,
            "Error on parsing online"
        )
        self.assertEqualTree(et.fromstring(xml_text), distribution.to_element(), "Creator error on to element")

    def test_parse_offline(self):
        xml_text = """
<distribution id="1" scope="document">
    <offline>
        <mediumName>Tape, 3.5 inch Floppy Disk, hardcopy</mediumName>
    </offline>
</distribution>
        """
        distribution = EMLDistribution.from_string(xml_text)
        self.assertEqual("1", distribution.id, "Error on parsing id")
        self.assertEqual(
            "Tape, 3.5 inch Floppy Disk, hardcopy",
            distribution.offline.medium_name,
            "Error on parsing offline"
        )
        self.assertEqualTree(et.fromstring(xml_text), distribution.to_element(), "Creator error on to element")

    def test_parse_referencing(self):
        xml_text = """
<distribution>
    <references>1</references>
</distribution>
        """
        distribution = EMLDistribution.from_string(xml_text)
        self.assertEqual("1", distribution.id, "Error on parsing id")
        self.assertTrue(
            distribution.referencing,
            "Error on parsing referencing"
        )
        self.assertEqualTree(et.fromstring(xml_text), distribution.to_element(), "Creator error on to element")


if __name__ == '__main__':
    unittest.main()



import os
import tempfile
import unittest
import zipfile

from dwca.base import DarwinCoreArchive
from eml import EML
from eml.types import ResponsibleParty, IndividualName
from test_dwca.test_dwca_common import TestDWCACommon

PATH = os.path.abspath(os.path.dirname(__file__))


class TestDWCA(TestDWCACommon):
    def test_minimal(self):
        super().__test_minimal__()

    def test_attributes(self):
        super().__test_attributes__()

    def test_merge(self):
        super().__test_merge__()

    def test_read_lazy(self):
        lazy_dwca = DarwinCoreArchive.from_file(os.path.join(PATH, os.pardir, "example_data", "example_archive.zip"), lazy=True)
        self.assertEqual(
            0, len(lazy_dwca.core.__entries__), "Entries populated in core."
        )
        self.assertNotEqual(
            0, len(lazy_dwca.core), "Laziness do not load length in core."
        )
        for extension in lazy_dwca.extensions:
            self.assertEqual(
                0, len(extension.__entries__), f"Entries populated in extension {extension.uri}."
            )
            self.assertNotEqual(
                0, len(extension), f"Laziness do not load length in extension {extension.uri}."
            )

    def test_as_pandas(self):
        df = self.object.core.as_pandas()
        self.assertEqual(163460, len(df), "Wrong number of rows")
        self.assertEqual(len(self.object.core.__fields__), len(df.columns), "Wrong number of fields")
        for extension, rows in zip(self.object.extensions, [153621, 98518, 7617]):
            df = extension.as_pandas()
            self.assertEqual(rows, len(df), f"Wrong number of rows at {extension.__class__.__name__}")
            self.assertEqual(len(extension.__fields__), len(df.columns), "Wrong number of fields")

    def test_set_pandas_core(self):
        df = self.object.core.as_pandas()
        self.assertEqual(163460, len(df), "Wrong number of rows")
        df = df[0:100]
        self.assertEqual(100, len(df), "Wrong number of rows set")
        self.object.core.pandas = df
        self.assertEqual(100, len(self.object.core), "Wrong number of rows set")
        for extension in self.object.extensions:
            self.assertGreaterEqual(100, len(extension), f"Wrong number of rows set at {extension.__class__.__name__}")

    def test_generate_eml(self):
        new_dwca = DarwinCoreArchive("Empty Darwin Core")
        new_dwca.generate_eml()
        eml = new_dwca.metadata
        self.assertIsInstance(
            eml, EML, "Wrong metadata generated"
        )
        self.assertEqual(
            "Empty Darwin Core", eml.package_id, "Wrong EML id"
        )

    def test_to_file(self):
        with tempfile.NamedTemporaryFile("wb") as file:
            self.object.to_file(file.name)
            with zipfile.ZipFile(file.name, "r") as zip_file:
                self.assertEqual(6, len(zip_file.namelist()), "Error writing DwC-A file.")
                self.assertCountEqual(
                    ["meta.xml", "taxon.txt", "identification.txt",
                     "reference.txt", "speciesprofile.txt", "eml.xml"],
                    zip_file.namelist(),
                    "Incorrect file saved"
                )
        empty_dwca = DarwinCoreArchive("Empty One")
        with tempfile.NamedTemporaryFile("wb") as file:
            empty_dwca.to_file(file.name)
            with zipfile.ZipFile(file.name, "r") as zip_file:
                self.assertEqual(1, len(zip_file.namelist()), "Error writing empty DwC-A file.")
                self.assertCountEqual(
                    ["meta.xml"],
                    zip_file.namelist(),
                    "Incorrect file saved"
                )
        meta_dwca = DarwinCoreArchive("With EML")
        meta_dwca.generate_eml()
        meta_dwca.metadata.initialize_resource(
            "Example for Darwin Core Archive",
            ResponsibleParty(
                _id="1",
                individual_name=IndividualName(
                    last_name="Doe",
                    first_name="Jane",
                    salutation="Ms"
                )
            ),
            contact=[ResponsibleParty(_id="1", referencing=True)]
        )
        with tempfile.NamedTemporaryFile("wb") as file:
            meta_dwca.to_file(file.name)
            with zipfile.ZipFile(file.name, "r") as zip_file:
                self.assertEqual(2, len(zip_file.namelist()), "Error writing DwC-A file with metadata.")
                self.assertCountEqual(
                    ["meta.xml", "eml.xml"],
                    zip_file.namelist(),
                    "Incorrect file saved"
                )

    def test_str(self):
        self.assertEqual(
            "Example Package [Core: http://rs.tdwg.org/dwc/terms/Taxon, Entries: 163460]",
            str(self.object),
            "Error on DwCA string."
        )

    def test_repr(self):
        self.assertEqual(
            "<Darwin Core Archive (Example Package [Core: http://rs.tdwg.org/dwc/terms/Taxon, Entries: 163460])>",
            repr(self.object),
            "Error on DwCA string representation."
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import re
from typing import List, Type, Any, Union, Set, Iterable
from warnings import warn

from dwca.classes import DataFile, DataFileType
from dwca.terms import Field, TaxonID, ScientificNameID, AcceptedNameUsageID, ParentNameUsageID, OriginalNameUsageID, \
    NameAccordingToID, NamePublishedInID, TaxonConceptID, ScientificName, AcceptedNameUsage, ParentNameUsage, \
    OriginalNameUsage, NameAccordingTo, NamePublishedIn, NamePublishedInYear, HigherClassification, Kingdom, Phylum, \
    DWCClass, Order, Superfamily, Family, Subfamily, Tribe, Subtribe, Genus, GenericName, Subgenus, \
    InfragenericEpithet, SpecificEpithet, InfraspecificEpithet, CultivarEpithet, TaxonRank, VerbatimTaxonRank, \
    ScientificNameAuthorship, VernacularName, NomenclaturalCode, TaxonomicStatus, NomenclaturalStatus, TaxonRemarks
from xml_common.utils import OptionalTqdm

try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import polars as pl
except ImportError:
    pl = None


class Taxon(DataFile):
    """
    A group of organisms considered by taxonomists to form a homogeneous unit.

    Parameters
    ----------
    _id : int
        Unique identifier for the core entity.
    files : str
        File location, in the archive, this is inside the `zip` file.
    fields : List[Field]
        A list of the Field (columns) in the Core data entity.
    data_file_type: DataFileType
        The Data File Type in the Darwin Core Archive.
    encoding : str, optional
        Encoding of the file location (`files` parameter), default is "utf-8".
    lines_terminated_by : str, optional
        Delimiter of lines on the file, default `"\\\\n"`.
    fields_terminated_by : str, optional
        Delimiter of the file (cells) on the file, default `","`.
    fields_enclosed_by : str, optional
        Specifies the character used to enclose (mark the start and end of) each field, default empty `""`.
    ignore_header_lines : int, optional
        Ignore headers at the start of document, can be one line or a list of them, default 0 (first line).
    """
    URI = DataFile.URI + "Taxon"
    __field_class__ = DataFile.__field_class__ + [
        TaxonID, ScientificNameID, AcceptedNameUsageID,
        ParentNameUsageID, OriginalNameUsageID,
        NameAccordingToID, NamePublishedInID, TaxonConceptID,
        ScientificName, AcceptedNameUsage, ParentNameUsage,
        OriginalNameUsage, NameAccordingTo, NamePublishedIn,
        NamePublishedInYear, HigherClassification, Kingdom,
        Phylum, DWCClass, Order, Superfamily, Family,
        Subfamily, Tribe, Subtribe, Genus, GenericName,
        Subgenus, InfragenericEpithet, SpecificEpithet,
        InfraspecificEpithet, CultivarEpithet, TaxonRank,
        VerbatimTaxonRank, ScientificNameAuthorship,
        VernacularName, NomenclaturalCode, TaxonomicStatus,
        NomenclaturalStatus, TaxonRemarks,
    ]

    def __init__(
            self, _id: int, files: str,
            fields: List[Field],
            data_file_type: DataFileType = DataFileType.CORE,
            encoding: str = "utf-8",
            lines_terminated_by: str = "\n",
            fields_terminated_by: str = ",",
            fields_enclosed_by: str = "",
            ignore_header_lines: int = 0,
    ) -> None:
        super().__init__(
            _id, files, fields, data_file_type, encoding,
            lines_terminated_by, fields_terminated_by,
            fields_enclosed_by, ignore_header_lines,
        )
        return

    def _filter_by_taxa_(
            self,
            taxa_field: Type[Field],
            taxa_name: str,
            taxa: List[str],
            filter_with_rank: bool = True,
            fuzzy_threshold: float = -1,
    ) -> None:
        assert taxa_field.URI in self.fields, f"{taxa_name} must be in fields of this class to use this feature."
        tqdm = OptionalTqdm(total=100)
        tqdm.set_descriptor(desc="Getting Taxon ID")
        if fuzzy_threshold > 0:
            try:
                from rapidfuzz import process
            except ImportError:
                raise ImportError("Install rapidfuzz to use this feature.")
        try:
            df = self.pandas
            if fuzzy_threshold < 0:
                mask = df[ScientificName.name_cls()].isin(taxa)
            else:
                tqdm.reset(total=len(df))
                def process_with_bar(x: str) -> float:
                    tqdm.update()
                    return process.extractOne(x, taxa)[1]
                matches = df[ScientificName.name_cls()].apply(process_with_bar)
                mask = matches >= fuzzy_threshold
            if filter_with_rank:
                mask &= df[TaxonRank.name_cls()].str.lower() == taxa_name.lower()
            taxa_id = df[mask][TaxonID.name_cls()]
        except ImportError:
            tqdm.reset(total=len(taxa))
            taxa_id = list()
            if filter_with_rank:
                for taxon in taxa:
                    taxon_found = self.__get_entry__(**{
                        ScientificName.name_cls(): taxon,
                        f"{TaxonRank.name_cls()}__case_insensitive": taxa_name
                    }, fuzzy_threshold=fuzzy_threshold)
                    if taxon_found is not None:
                        taxa_id.append(getattr(taxon_found, TaxonID.name_cls()))
                    tqdm.update()
            else:
                for taxon in taxa:
                    taxon_found = self.__get_entry__(**{ScientificName.name_cls(): taxon}, fuzzy_threshold=fuzzy_threshold)
                    if taxon_found is not None:
                        taxa_id.append(getattr(taxon_found, TaxonID.name_cls()))
                    tqdm.update()
            tqdm.reset(total=100)
        tqdm.update(n=10)
        postfix = {"Exact match found": len(taxa_id)}
        tqdm.set_postfix(ordered_dict=postfix)
        tqdm.set_descriptor(desc="Getting synonyms")
        complete_taxa = self.all_synonyms(taxa_id, get_names=True)
        postfix["Synonyms found"] = len(complete_taxa) - len(taxa_id)
        taxa_id = self.all_synonyms(taxa_id)
        tqdm.set_postfix(ordered_dict=postfix)
        tqdm.update(n=40)
        tqdm.set_descriptor(desc="Getting parents")
        parents = self.get_parents(taxa_id)
        parents.update(self.all_synonyms(parents))
        postfix["Parents found"] = len(parents)
        tqdm.set_postfix(ordered_dict=postfix)
        tqdm.update(n=40)
        tqdm.set_descriptor(desc=f"Filtering {taxa_name}")
        try:
            df = self.pandas
            name = taxa_field.name_cls()
            mask = df[TaxonID.name_cls()].isin(parents) | df[TaxonID.name_cls()].isin(taxa_id)
            if filter_with_rank:
                mask |= df[name].isin(complete_taxa)
            df = df[mask]
            self.pandas = df
        except ImportError:
            if filter_with_rank:
                def filter_taxa(entry: DataFile.Entry) -> bool:
                    return (getattr(entry, taxa_field.name_cls()) in complete_taxa or
                            getattr(entry, TaxonID.name_cls()) in parents or
                            getattr(entry, TaxonID.name_cls()) in taxa_id)
            else:
                def filter_taxa(entry: DataFile.Entry) -> bool:
                    return (getattr(entry, TaxonID.name_cls()) in parents or
                            getattr(entry, TaxonID.name_cls()) in taxa_id)
            self.__entries__ = list(filter(filter_taxa, self.__entries__))
        postfix["Total filtered"] = len(self)
        tqdm.set_postfix(ordered_dict=postfix)
        tqdm.update(n=10)
        tqdm.close()
        return

    def filter_by_kingdom(self, kingdoms: List[str], fuzzy_threshold: float = -1) -> None:
        """
        Filter data by a valid kingdoms.

        Parameters
        ----------
        kingdoms : List[str]
            Kingdom names to filter data.
        fuzzy_threshold : float, optional
            If given any value > 0 it will use Levenshtein Distance with that threshold instead of exact match.
        """
        return self._filter_by_taxa_(Kingdom, "Kingdom", kingdoms, fuzzy_threshold=fuzzy_threshold)

    def filter_by_phylum(self, phyla: List[str], fuzzy_threshold: float = -1) -> None:
        """
        Filter data by a valid phylum.

        Parameters
        ----------
        phyla : List[str]
            Phylum names to filter data.
        fuzzy_threshold : float, optional
            If given any value > 0 it will use Levenshtein Distance with that threshold instead of exact match.
        """
        return self._filter_by_taxa_(Phylum, "Phylum", phyla, fuzzy_threshold=fuzzy_threshold)

    def filter_by_class(self, classes: List[str], fuzzy_threshold: float = -1) -> None:
        """
        Filter data by a valid class.

        Parameters
        ----------
        classes : List[str]
            Class names to filter data.
        fuzzy_threshold : float, optional
            If given any value > 0 it will use Levenshtein Distance with that threshold instead of exact match.
        """
        return self._filter_by_taxa_(DWCClass, "Class", classes, fuzzy_threshold=fuzzy_threshold)

    def filter_by_order(self, orders: List[str], fuzzy_threshold: float = -1) -> None:
        """
        Filter data by a valid order.

        Parameters
        ----------
        orders : List[str]
            Order names to filter data.
        fuzzy_threshold : float, optional
            If given any value > 0 it will use Levenshtein Distance with that threshold instead of exact match.
        """
        return self._filter_by_taxa_(Order, "Order", orders, fuzzy_threshold=fuzzy_threshold)

    def filter_by_family(self, families: List[str], fuzzy_threshold: float = -1) -> None:
        """
        Filter data by a valid family.

        Parameters
        ----------
        families : List[str]
            Family names to filter data.
        fuzzy_threshold : float, optional
            If given any value > 0 it will use Levenshtein Distance with that threshold instead of exact match.
        """
        return self._filter_by_taxa_(Family, "Family", families, fuzzy_threshold=fuzzy_threshold)

    def filter_by_genus(self, genera: List[str], fuzzy_threshold: float = -1) -> None:
        """
        Filter data by a valid genus.

        Parameters
        ----------
        genera : List[str]
            Class names to filter genus.
        fuzzy_threshold : float, optional
            If given any value > 0 it will use Levenshtein Distance with that threshold instead of exact match.
        """
        return self._filter_by_taxa_(Genus, "Genus", genera, fuzzy_threshold=fuzzy_threshold)

    def filter_by_species(self, species: List[str], fuzzy_threshold: float = -1) -> None:
        """
        Filer data by species or any rank taxonomy below (subspecies, variety, form, etc.).

        In contrast with the other `filter_by_` taxonomy methods, this one filter
        the taxonomic data using the scientific name field :class:`dwca.terms.taxon.ScientificName`.

        .. warning:: Because of that, use this method with precautions.
           If a scientific name of a rank above species (genus, order,
           etc...) is used, it could result in unexpected behaviour.

        Parameters
        ----------
        species : List[str]
            Scientific Name of species (or rank below) to filter data.
        fuzzy_threshold : float, optional
            If given any value > 0 it will use Levenshtein Distance with that threshold instead of exact match.
        """
        return self._filter_by_taxa_(ScientificName, "Species", species, filter_with_rank=False, fuzzy_threshold=fuzzy_threshold)

    def get_parents(self, taxa_id: List[str]) -> Set[str]:
        """
        Get a list of taxa ids of the parent of the list taxa id provided.

        Parameters
        ----------
        taxa_id : List[str]
            A list of :class:`dwca.terms.taxon.TaxonID` to look for parents.

        Returns
        -------
        Set[str]
            Set of taxa ids.
        """
        parent_taxa = set()
        try:
            df = self.pandas
            current_taxa = self.__get_rows__(taxa_id)
            while len(current_taxa) > 0:
                new_found_parents = current_taxa[ParentNameUsageID.name_cls()]
                new_found_parents = new_found_parents[
                    pd.notna(new_found_parents) &
                    (new_found_parents != "")
                ]
                parent_taxa.update(list(new_found_parents))
                current_taxa = df[df[TaxonID.name_cls()].isin(new_found_parents)]
        except ImportError:
            entries = self.__get_rows__(taxa_id)
            current_taxa = entries.copy()
            while len(current_taxa) > 0:
                new_found_parents = [getattr(current, ParentNameUsageID.name_cls()) for current in current_taxa]
                new_found_parents = list(filter(
                    lambda current: current != "" and current is not None,
                    new_found_parents
                ))
                parent_taxa.update(new_found_parents)
                current_taxa = self.__get_entries__(**{f"{TaxonID.name_cls()}__isin": new_found_parents})
        return parent_taxa

    def all_synonyms(self, taxa_id: Iterable[str], get_names: bool = False) -> List[str]:
        """
        Get a list of all valid names of a list of taxa.

        Parameters
        ----------
        taxa_id : Iterable[str]
            A list (or iterable) of :class:`dwca.terms.taxon.TaxonID` value.
        get_names : bool
            Whether to get :class:`dwca.terms.taxon.ScientificName` or :class:`dwca.terms.taxon.TaxonID`.

        Returns
        -------
        List[str]
            A list of :class:`dwca.terms.taxon.TaxonID`.
        """
        try:
            df = self.pandas
            current_taxa = self.__get_rows__(taxa_id)
            accepted_names = current_taxa[AcceptedNameUsageID.name_cls()]
            return list(
                df[df[AcceptedNameUsageID.name_cls()].isin(accepted_names)][
                    ScientificName.name_cls() if get_names else TaxonID.name_cls()
                ]
            )
        except ImportError:
            entries = self.__get_rows__(taxa_id)
            accepted_names = [getattr(entry, AcceptedNameUsageID.name_cls()) for entry in entries]
            synonyms = self.__get_entries__(**{f"{AcceptedNameUsageID.name_cls()}__isin": accepted_names})
            field = ScientificName.name_cls() if get_names else TaxonID.name_cls()
            return [getattr(synonymous, field) for synonymous in synonyms]

    def __get_rows__(self, taxa_id: Iterable[str]) -> Union[pd.DataFrame, List[DataFile.Entry]]:
        try:
            df = self.pandas
            current_taxa = df[df[TaxonID.name_cls()].isin(taxa_id)]
            if len(current_taxa) != len(taxa_id):
                candidates = pd.Series(taxa_id)
                not_present = candidates[~candidates.isin(df[TaxonID.name_cls()])]
                warn(f"{', '.join(list(not_present))} not found in data file.", category=RuntimeWarning)
            return current_taxa
        except ImportError:
            entries = self.__get_entries__(**{f"{TaxonID.name_cls()}__isin": taxa_id})
            if len(entries) != len(taxa_id):
                found_taxa = [getattr(entry, TaxonID.name_cls()) for entry in entries]
                not_present = list(filter(lambda candid: candid not in found_taxa, taxa_id))
                warn(f"{', '.join(not_present)} not found in data file.", category=RuntimeWarning)
            return entries

    def __get_entry__(self, fuzzy_threshold: float =-1, **kwargs) -> DataFile.Entry | None:
        for candid in self.__entries__:
            found = True
            for key, value in kwargs.items():
                found &= Taxon.__compare__(candid, key, value, fuzzy_threshold=fuzzy_threshold)
                if not found:
                    break
            if found:
                return candid
        return None

    def __get_entries__(self, fuzzy_threshold: float =-1, **kwargs) -> List[DataFile.Entry]:
        results = list()
        for candid in self.__entries__:
            found = True
            for key, value in kwargs.items():
                found &= Taxon.__compare__(candid, key, value, fuzzy_threshold=fuzzy_threshold)
                if not found:
                    break
            if found:
                results.append(candid)
        return results

    @staticmethod
    def __compare__(candid: DataFile.Entry, key: str, value: Any, fuzzy_threshold: float = -1) -> bool:
        match = re.match(r"(\w+)__case_insensitive$", key)
        if match:
            if fuzzy_threshold > 0:
                from rapidfuzz import fuzz
                ratio = fuzz.ratio(getattr(candid, match.group(1)).lower(), value.lower())
                return ratio >= fuzzy_threshold
            return getattr(candid, match.group(1)).lower() == value.lower()
        match = re.match(r"(\w+)__isin$", key)
        if match:
            if fuzzy_threshold > 0:
                from rapidfuzz import process
                _, ratio = process.extractOne(getattr(candid, match.group(1)), value)
                return ratio >= fuzzy_threshold
            return getattr(candid, match.group(1)) in value
        if fuzzy_threshold > 0:
            from rapidfuzz import fuzz
            ratio = fuzz.ratio(getattr(candid, key), value)
            if getattr(candid, key) == "Rtfkiaicpdng (Abifwvxqn) gurqtwpof f. prczacpvtdtu 'xzhgezqpaorp'":
                import sys
                print(candid, key, value, ratio, file=sys.stderr)
            return ratio >= fuzzy_threshold
        return getattr(candid, key) == value



from unittest import main
from lxml import etree as et

from eml.resources import EMLDataset
from eml.types import Methods, ProcedureStep
from test_xml.test_xml import TestXML


class TestMethods(TestXML):
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
        "scope": "document"
    }

    def test_parse_none(self):
        self.assertIsNone(Methods.parse(None, {}), "Methods from nowhere.")

    def test_parse_empty(self):
        text_xml = """
<methods>
</methods>
        """
        with self.assertRaisesRegex(TypeError, "At least one step must be given."):
            Methods.from_string(text_xml)

    def test_parse_simple(self):
        text_xml = """
<methods>
    <methodStep>
        <description>
            <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
        </description>
    </methodStep>
</methods>
        """
        methods = Methods.from_string(text_xml)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            methods.method_steps[0].description.paragraphs[0],
            "Error parsing method step."
        )
        self.assertIsNone(methods.sampling, "Sampling method from nowhere.")
        self.assertEqual(0, len(methods.quality_control), "Quality control from nowhere.")
        self.assertEqualTree(et.fromstring(text_xml), methods.to_element(), "Error on to element.")

    def test_parse_data_source(self):
        text_xml = """
<methods>
    <methodStep>
        <description>
            <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
        </description>
        <dataSource>
            <references>dataset-1</references>
        </dataSource>
    </methodStep>
</methods>
        """
        methods = Methods.from_string(text_xml)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            methods.method_steps[0].description.paragraphs[0],
            "Error parsing method step."
        )
        self.assertIsNone(methods.sampling, "Sampling method from nowhere.")
        self.assertEqual(0, len(methods.quality_control), "Quality control from nowhere.")
        self.assertEqualTree(et.fromstring(text_xml), methods.to_element(), "Error on to element.")

    def test_parse_data_source_complex(self):
        text_xml = """
<methods>
    <methodStep>
        <description>
            <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit.</para>
        </description>
        <dataSource>
            <references>dataset-1</references>
        </dataSource>
    </methodStep>
    <methodStep>
        <description>
            <para>Etiam vitae diam nec eros interdum posuere.</para>
        </description>
    </methodStep>
</methods>
        """
        methods = Methods.from_string(text_xml)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit.",
            methods.method_steps[0].description.paragraphs[0],
            "Error parsing method step."
        )
        self.assertIsNone(methods.sampling, "Sampling method from nowhere.")
        self.assertEqual(0, len(methods.quality_control), "Quality control from nowhere.")
        self.assertEqualTree(et.fromstring(text_xml), methods.to_element(), "Error on to element.")

    def test_incomplete_data_source(self):
        methods = Methods(
            method_steps=[
                ProcedureStep.from_string("<methodStep><description>"
                                          "<para>Lorem ipsum dolor sit amet, consectetur adipiscing elit.</para>"
                                          "</description></methodStep>"),
                ProcedureStep.from_string("<methodStep><description>"
                                          "<para>Etiam vitae diam nec eros interdum posuere.</para>"
                                          "</description></methodStep>"),
            ],
            data_sources=[
                [EMLDataset(_id="dataset-1", referencing=True)],
            ]
        )
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit.",
            methods.method_steps[0].description.paragraphs[0],
            "Error initializing methods."
        )
        self.assertEqual(
            0, len(methods.method_steps[1].data_source),
            "Data source from nowhere on second step."
        )

    def test_parse_method_step(self):
        text_xml = """
<methodStep>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
    </description>
</methodStep>
        """
        method_step = Methods.MethodStep.from_string(text_xml)
        self.assertEqualTree(et.fromstring(text_xml), method_step.to_element(), "Error on step method to element.")

    def test_parse_sampling(self):
        text_xml = """
<methods>
    <methodStep>
        <description>
            <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
        </description>
    </methodStep>
    <sampling>
    </sampling>
</methods>
        """
        methods = Methods.from_string(text_xml)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            methods.method_steps[0].description.paragraphs[0],
            "Error parsing method step."
        )
        # self.assertIsNone(methods.sampling, "Sampling method from nowhere.")
        # self.assertEqual(0, len(methods.quality_control), "Quality control from nowhere.")
        # self.assertEqualTree(et.fromstring(text_xml), methods.to_element(), "Error on to element.")


if __name__ == '__main__':
    main()



from xml_common.utils import CamelCaseEnum


class LengthUnit(CamelCaseEnum):
    """
    The list of units which are of length type, or have a parentSI of meter SI.

    Attributes
    ----------
    METER : float
        International meter.
    NANOMETER : float
        International nanometer.
    MICROMETER : float
        International micrometer.
    MICRON : float
        Non International System of Units (SI) name of micrometer.
    MILLIMETER : float
        International millimeter.
    CENTIMETER : float
        International centimeter.
    DECIMETER : float
        International decimeter.
    DEKAMETER : float
        International dekameter.
    HECTOMETER : float
        International hectometer.
    KILOMETER : float
        International kilometer.
    MEGAMETER : float
        International megameter.
    ANGSTROM : float
        Ångström metric unit.
    INCH : float
         British imperial and United States customary inch.
    FOOT_US : float
        United States foot.
    FOOT : float
        International foot.
    FOOT_GOLD_COAST : float
        Gold Coast foot.
    FATHOM : float
        A measure (especially of depth) of six feet.
    NAUTICALMILE : float
        Nautical mile.
    YARD : float
        International yard.
    YARD_INDIAN : float
        Indian yard.
    LINK_CLARKE : float
        Link (Clarke's ratio).
    YARD_SEARS : float
        Yard (Sears).
    MILE : float
        International mile.
    """
    METER = 1e0
    NANOMETER = 1e-9
    MICROMETER = 1e-6
    MICRON = MICROMETER
    MILLIMETER = 1e-3
    CENTIMETER = 1e-2
    DECIMETER = 1e-1
    DEKAMETER = 1e1
    HECTOMETER = 1e2
    KILOMETER = 1e3
    MEGAMETER = 1e6
    ANGSTROM = 1e-10
    INCH = 0.0254
    FOOT_US = 0.3048006096012192
    FOOT = 0.3048
    FOOT_GOLD_COAST = 0.30479971018150881758
    FATHOM = 1.8288
    NAUTICALMILE = 1852
    YARD = 0.9144
    YARD_INDIAN = 0.9143985307444408
    LINK_CLARKE = 0.2011661949
    YARD_SEARS = 0.9143984146160287
    MILE = 1609.347218694438

    @property
    def valid_name(self) -> str:
        """
        Generate the valid name of the unit.

        Returns
        -------
        str
            Name of the unit.
        """
        return "_".join([word.lower().capitalize() for word in self.name.split("_")])



from __future__ import annotations

from abc import ABC
from typing import Any, Dict

from lxml import etree as et

from xml_common.utils import format_to_type, unformat_type, type_to_sql
from xml_common import XMLObject


class Field(XMLObject, ABC):
    """
    Element use to specify the location and content of data within a :class:`dwca.classes.data_file.DataFile`.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    PRINCIPAL_TAG = "field"
    URI = "http://rs.tdwg.org/dwc/terms/"
    """str: URI for the term represented by this field."""
    TYPE = Any
    """type: Type of the field."""

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__()
        self.__index__ = int(index)
        self.__default__ = default
        self.__vocabulary__ = vocabulary
        return

    @property
    def index(self) -> int:
        """int: Specifies the position of the column in the row."""
        return self.__index__

    @index.setter
    def index(self, index: int) -> None:
        self.__index__ = index
        return

    @property
    def default(self) -> Any:
        """Any: Specifies a value to use if one is not supplied."""
        return self.__default__

    @property
    def vocabulary(self) -> str:
        """str: An URI for a vocabulary that the source values for this Field are based on."""
        return self.__vocabulary__

    @property
    def uri(self) -> str:
        """str: An URI for the term represented by this field."""
        return self.URI

    @classmethod
    def name_cls(cls) -> str:
        """str: The name of the field."""
        names = cls.URI.split("/")
        if len(names) > 1:
            return names[-1]
        else:
            return names[0]

    @property
    def name(self) -> str:
        """str: The name of the field."""
        names = self.uri.split("/")
        if len(names) > 1:
            return names[-1]
        else:
            return names[0]

    @property
    def sql_type(self) -> str:
        """str: Type of field in a SQL relational database."""
        return type_to_sql(self.TYPE)

    def format(self, value: str) -> TYPE:
        """
        Format value in the TYPE of the field.

        Parameters
        ----------
        value : str
            Value to be formatted in the type of the respected field.

        Returns
        -------
        TYPE
            Formatted value.
        """
        try:
            return format_to_type(value, self.TYPE)
        except Exception as e:
            raise type(e)(f"{e} Must overwrite `Field.format` method to use this field.")

    def unformat(self, value: TYPE) -> str:
        """
        Encode value from TYPE to a string.

        Parameters
        ----------
        value : TYPE
            Value to be encoded.

        Returns
        -------
        str
            Text encoded value..
        """
        return unformat_type(value, self.TYPE)

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> Field | None:
        """
        Generate a Field object from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance.
        nmap : Dict
            Namespace.

        Returns
        -------
        Field
            A Field object.
        """
        if element is None:
            return None
        if "term" not in element.attrib:
            raise TypeError("Field must have a term")
        field = cls(
            index=element.get("index", None),
            default=element.get("default", None),
            vocabulary=element.get("vocabulary", None)
        )
        if field.__default__ is not None:
            field.__default__ = field.format(field.__default__)
        field.__namespace__ = nmap
        return field

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance from this object.

        Returns
        -------
        lxml.etree.Element
            An XML element instance.
        """
        element = super().to_element()
        element.set("term", self.uri)
        if self.__index__ is not None:
            element.set("index", str(self.__index__))
        if self.__default__ is not None:
            element.set("default", self.unformat(self.__default__))
        if self.__vocabulary__ is not None:
            element.set("vocabulary", self.__vocabulary__)
        return element

    def __repr__(self) -> str:
        return f"<Field [term={self.uri}]>"



import unittest
from typing import List, Tuple, Any, Union

import datetime as dt

from datetime_interval import Interval

from xml_common.utils import format_to_type


class MetaNewType(type):
    def __str__(cls) -> str:
        return 'NewType'


class MetaAnotherType(type):
    def __str__(cls) -> str:
        return '<AnotherType>'


class NewType(metaclass=MetaNewType):
    pass


class AnotherType(metaclass=MetaAnotherType):
    pass


class FormatTest(unittest.TestCase):

    def test_format_int(self):
        a = "23"
        self.assertEqual(23, format_to_type(a, int), "Incorrect format to integer.")

    def test_format_float(self):
        a = "23"
        self.assertGreaterEqual(1e-10, abs(23 - format_to_type(a, float)), "Incorrect format to float.")
        b = "3.14"
        self.assertGreaterEqual(1e-10, abs(3.14 - format_to_type(b, float)), "Incorrect format to float.")

    def test_format_str(self):
        a = "23"
        self.assertEqual("23", format_to_type(a, str), "Incorrect format to string.")
        b = "abc"
        self.assertEqual("abc", format_to_type(b, str), "Incorrect format to string.")

    def test_format_list(self):
        a = "1 | 2 | 3"
        self.assertEqual(["1", "2", "3"], format_to_type(a, List[str]), "Incorrect format to list of string.")

    def test_format_tuple(self):
        a = "1/2/3"
        self.assertEqual(
            ("1", "2", "3"),
            format_to_type(a, Tuple[str, str, str]),
            "Incorrect format to tuple of string."
        )
        self.assertEqual(
            (1, 2, 3),
            format_to_type(a, Tuple[int, int, int]),
            "Incorrect format to tuple of integer."
        )
        self.assertEqual(
            (1, "2", 3),
            format_to_type(a, Tuple[int, str, int]),
            "Incorrect format to tuple."
        )

    def test_address_error(self):
        a = "abc"
        self.assertIsNone(format_to_type(a, int, True), "Format incorrect not addressed")

    def test_not_address_error(self):
        a = "abc"
        self.assertRaises(
            ValueError,
            format_to_type,
            a, int, False
        )

    def test_not_type_error(self):
        a = "abc"
        self.assertRaisesRegex(TypeError, "<NewType>",
                               format_to_type, a, NewType)
        self.assertRaisesRegex(TypeError, "<AnotherType>",
                               format_to_type, a, AnotherType)

    def test_warning_any(self):
        a = "abc"
        with self.assertWarnsRegex(UserWarning, "<Any>"):
            self.assertEqual(
                a, format_to_type(a, Any), "Error on format Any"
            )

    def test_format_union(self):
        a = "abc"
        self.assertEqual("abc", format_to_type(a, Union[int, str]), "Error on format to string (in union)")

    def test_format_union_error(self):
        a = "abc"
        self.assertRaisesRegex(
            TypeError,
            "any of <class 'int'>, <class 'float'>",
            format_to_type,
            a, Union[int, float]
        )

    def test_format_datetime(self):
        a = "2024-04-05T01:01:07-0600"
        self.assertEqual(
            dt.datetime(2024, 4, 5, 1, 1, 7, tzinfo=dt.timezone(dt.timedelta(hours=-6))),
            format_to_type(a, dt.datetime),
            "Incorrect format of datetime `%Y-%m-%dT%H:%M:%S%z`"
        )
        b = "2024-04-04T01:07-0500"
        self.assertEqual(
            dt.datetime(2024, 4, 4, 1, 7, 0, tzinfo=dt.timezone(dt.timedelta(hours=-5))),
            format_to_type(b, dt.datetime),
            "Incorrect format of datetime `%Y-%m-%dT%H:%M%z`"
        )
        c = "2024"
        self.assertEqual(
            dt.datetime(2024, 1, 1, 0, 0, 0),
            format_to_type(c, dt.datetime),
            "Incorrect format of datetime `%Y`"
        )
        d = ""
        self.assertIsNone(
            format_to_type(d, dt.datetime),
            "Incorrect format of empty datetime"
        )

    def test_format_datetime_error(self):
        a = "2024/04/05<->01:01:07"
        self.assertRaisesRegex(
            ValueError,
            "does not match",
            format_to_type,
            a, dt.datetime
        )

    def test_format_interval(self):
        a = "2024-04-04T01:01:07-0600/2024-04-05T01:07-0500"
        self.assertEqual(
            dt.datetime(2024, 4, 4, 1, 1, 7, tzinfo=dt.timezone(dt.timedelta(hours=-6))),
            format_to_type(a, Interval).start,
            "Incorrect format of datetime `%Y-%m-%dT%H:%M:%S%z`"
        )
        self.assertEqual(
            dt.datetime(2024, 4, 5, 1, 7, 0, tzinfo=dt.timezone(dt.timedelta(hours=-5))),
            format_to_type(a, Interval).end,
            "Incorrect format of datetime `%Y-%m-%dT%H:%M%z`"
        )
        b = "2023/2024-04"
        self.assertEqual(
            dt.datetime(2023, 1, 1),
            format_to_type(b, Interval).start,
            "Incorrect format of datetime `%Y`"
        )
        self.assertEqual(
            dt.datetime(2024, 4, 1),
            format_to_type(b, Interval).end,
            "Incorrect format of datetime `%Y-%m`"
        )


if __name__ == '__main__':
    unittest.main()



import unittest
from lxml import etree as et

from xml_common.utils import Language
from eml.types import EMLAddress
from test_xml.test_xml import TestXML


class TestEMLAddress(TestXML):
    DEFAULT_TAGS = {
        "scope": "document",
    }

    EXAMPLE_XML = """
<address id="1" scope="system" system="https://ieb-chile.cl">
    <deliveryPoint xml:lang="esp">Calle Las Palmeras #3425</deliveryPoint>
    <city xml:lang="esp">Ñuñoa</city>
    <administrativeArea xml:lang="esp">Región Metropolitana</administrativeArea>
    <postalCode xml:lang="esp">7800003</postalCode>
    <country xml:lang="esp">Chile</country>
</address>
    """

    def test_eml_address(self):
        address = EMLAddress(
            delivery_point="7209 Coast Drive, Building 44",
            city="San Francisco",
            administrative_area="Colorado",
            postal_code="93106-2231",
            country="U.S.A.",
        )
        self.assertEqual("7209 Coast Drive, Building 44", address.delivery_point[0], "Incorrect delivery point")
        self.assertEqual(Language.ENG, address.delivery_point[0].language, "Incorrect language of delivery point")
        self.assertEqual("San Francisco", address.city, "Incorrect city")
        self.assertEqual(Language.ENG, address.city.language, "Incorrect language of city")
        self.assertEqual("Colorado", address.administrative_area, "Incorrect administrative area")
        self.assertEqual(
            Language.ENG,
            address.administrative_area.language,
            "Incorrect language of administrative area"
        )
        self.assertEqual("93106-2231", address.postal_code, "Incorrect postal_code")
        self.assertEqual(Language.ENG, address.postal_code.language, "Incorrect language of postal_code")
        self.assertEqual("U.S.A.", address.country, "Incorrect country")
        self.assertEqual(Language.ENG, address.country.language, "Incorrect language of country")

    def test_eml_address_spanish(self):
        address = EMLAddress(
            delivery_point="Calle Las Palmeras #3425",
            city="Ñuñoa",
            administrative_area="Región Metropolitana",
            postal_code="7800003",
            country="Chile",
            language=Language.ESP,
        )
        self.assertEqual("Calle Las Palmeras #3425", address.delivery_point[0], "Incorrect delivery point")
        self.assertEqual(Language.ESP, address.delivery_point[0].language, "Incorrect language of delivery point")
        self.assertEqual("Ñuñoa", address.city, "Incorrect city")
        self.assertEqual(Language.ESP, address.city.language, "Incorrect language of city")
        self.assertEqual("Región Metropolitana", address.administrative_area, "Incorrect administrative area")
        self.assertEqual(
            Language.ESP,
            address.administrative_area.language,
            "Incorrect language of administrative area"
        )
        self.assertEqual("7800003", address.postal_code, "Incorrect postal_code")
        self.assertEqual(Language.ESP, address.postal_code.language, "Incorrect language of postal_code")
        self.assertEqual("Chile", address.country, "Incorrect country")
        self.assertEqual(Language.ESP, address.country.language, "Incorrect language of country")

    def test_parse_none(self):
        self.assertIsNone(EMLAddress.parse(None, {}), "Object from None")

    def test_parse(self):
        address = EMLAddress.from_string(self.EXAMPLE_XML)
        self.assertEqual("Calle Las Palmeras #3425", address.delivery_point[0], "Incorrectly set delivery point")
        self.assertEqual("Ñuñoa", address.city, "Incorrectly set city")
        self.assertEqual(
            "Región Metropolitana",
            address.administrative_area,
            "Incorrectly set city administrative area"
        )
        self.assertEqual("7800003", address.postal_code, "Incorrectly set postal code")
        self.assertEqual("Chile", address.country, "Incorrectly set country")
        self.assertEqualTree(et.fromstring(self.EXAMPLE_XML), address.to_element(), "Error on to element")

    def test_mode_delivery_point(self):
        example_xml = self.EXAMPLE_XML.replace(
            '<deliveryPoint xml:lang="esp">Calle Las Palmeras #3425</deliveryPoint>',
            """
<deliveryPoint
    xml:lang="esp"
>Calle Las Palmeras #3425</deliveryPoint>
<deliveryPoint
    xml:lang="eng"
>Street Las Palmeras #3425</deliveryPoint>
            """
        )
        address = EMLAddress.from_string(example_xml)
        self.assertEqual(2, len(address.delivery_point), "Incorrectly set delivery point")
        self.assertEqual(
            "Calle Las Palmeras #3425",
            address.delivery_point[0],
            "Incorrectly set spanish delivery point"
        )
        self.assertEqual(
            "Street Las Palmeras #3425",
            address.delivery_point[1],
            "Incorrectly set english delivery point"
        )
        self.assertEqual(
            Language.ESP,
            address.delivery_point[0].language,
            "Incorrectly set spanish delivery point language"
        )
        self.assertEqual(
            Language.ENG,
            address.delivery_point[1].language,
            "Incorrectly set english delivery point language"
        )
        self.assertEqualTree(et.fromstring(example_xml), address.to_element(), "Error on to element")

    def test_wrong_scope(self):
        example_xml = self.EXAMPLE_XML.replace('scope="system"', 'scope="invalidScope"')
        self.assertRaises(ValueError, EMLAddress.from_string, example_xml)

    def test_reference(self):
        reference_xml = """
<address>
    <references>1</references>
</address>
        """
        eml_address = EMLAddress.from_string(reference_xml)
        self.assertEqual("1", eml_address.id, "Id set incorrectly")
        self.assertTrue(eml_address.referencing, "Referencing wrong initialized")
        self.assertIsNone(
            eml_address.references.system,
            "References system wrong initialized"
        )
        self.assertEqualTree(
            et.fromstring(reference_xml),
            eml_address.to_element(),
            "Wrong referrer to element"
        )

    def test_reference_system(self):
        reference_xml = """
<address>
    <references system="http://gbif.org">1</references>
</address>
        """
        eml_address = EMLAddress.from_string(reference_xml)
        self.assertEqual("1", eml_address.id, "Id set incorrectly")
        self.assertTrue(eml_address.referencing, "Referencing wrong initialized")
        self.assertEqual(
            "http://gbif.org",
            eml_address.references.system,
            "References system wrong initialized"
        )
        self.assertEqualTree(
            et.fromstring(reference_xml),
            eml_address.to_element(),
            "Wrong referrer to element"
        )

    def test_invalid_reference(self):
        reference_xml = """
<address id="10">
    <references>1</references>
</address>
        """
        self.assertRaises(
            AssertionError,
            EMLAddress.from_string,
            reference_xml
        )

    # TODO: Test different ways to equal on Address


if __name__ == '__main__':
    unittest.main()



import os
import unittest

from dwca.classes import OutsideClass, Taxon
from test_dwca_classes.test_outside_class_common import TestOutsideCommon
from test_dwca_classes.test_taxon_common import TestTaxonCommon
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestDataFile(TestXML):
    def test_assertion(self):
        _, xml_file_1, text_1 = TestTaxonCommon.read_xml(os.path.join(
            PATH, os.pardir, "example_data", "dataset1_meta.xml"
        ))
        xml_file_2, text_2 = TestOutsideCommon.read_xml(os.path.join(
            PATH, os.pardir, "example_data", "dataset1_meta.xml"
        ))
        taxon1 = Taxon.from_string(text_1)
        with open(os.path.join(PATH, os.pardir, "example_data", "taxon1.txt"), "r", encoding="utf-8") as taxa_file:
            taxon1.read_file(taxa_file.read())
        extension = OutsideClass.from_string(text_2)
        with open(
                os.path.join(PATH, os.pardir, "example_data", "extension_example1.txt"),
                "r", encoding="utf-8"
        ) as file:
            extension.read_file(file.read())
        self.assertRaises(AssertionError, taxon1.merge, extension)
        self.assertRaises(AssertionError, extension.merge, taxon1)


if __name__ == '__main__':
    unittest.main()



import unittest
import datetime as dt
import lxml.etree as et

from eml.types import ChangeHistory
from test_xml.test_xml import TestXML


class TestChangeHistory(TestXML):
    def test_parse_none(self):
        self.assertIsNone(ChangeHistory.parse(None, {}), "Change History from None.")

    def test_parse_simple(self):
        text_xml = """
<changeHistory>
    <changeScope>Scope</changeScope>
    <oldValue>Document</oldValue>
    <changeDate>2001</changeDate>
</changeHistory>
        """
        change_history = ChangeHistory.from_string(text_xml)
        self.assertEqual("Scope", change_history.change_scope, "Error on parsing change scope.")
        self.assertEqual("Document", change_history.old_value, "Error on parsing old value.")
        self.assertEqual(dt.date(year=2001, month=1, day=1), change_history.change_date, "Error on parsing change date.")
        self.assertEqualTree(et.fromstring(text_xml), change_history.to_element(), "Error on to element.")

    def test_parse_with_comment(self):
        text_xml = """
<changeHistory>
    <changeScope>Scope</changeScope>
    <oldValue>Document</oldValue>
    <changeDate>2001</changeDate>
    <comment>Testing change history</comment>
</changeHistory>
        """
        change_history = ChangeHistory.from_string(text_xml)
        self.assertEqual("Scope", change_history.change_scope, "Error on parsing change scope.")
        self.assertEqual("Document", change_history.old_value, "Error on parsing old value.")
        self.assertEqual(dt.date(year=2001, month=1, day=1), change_history.change_date,
                         "Error on parsing change date.")
        self.assertEqualTree(et.fromstring(text_xml), change_history.to_element(), "Error on to element.")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from abc import ABC
from typing import Union, List, Dict, Tuple
from warnings import warn

import datetime as dt

from lxml import etree as et

from xml_common.utils import Language, format_datetime
from eml.resources import EMLKeywordSet, EMLLicense, EMLDistribution, EMLCoverage
from eml.types import I18nString, EMLObject, Scope, ExtensionString, ResponsibleParty, EMLTextType, SemanticAnnotation, \
    Role


class Resource(EMLObject, ABC):
    """
    Abstract class representing any resources on an EML document

    Parameters
    ----------
    _id : str, optional
        Unique identifier
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined
    references_system : str, optional
        System attribute of reference
    alternative_identifier : List[ExtendedString], optional
        A list of secondary identifier for this entity.
    short_name : str, optional
        A short name that describes the resource, sometimes a filename.
    titles : List[I18nString or str]
        A brief description of the resource. At least one must be given.
    creators : List[ResponsibleParty]
        The people or organizations who created this resource. At least one must be given.
    metadata_providers : List[ResponsibleParty], optional
        The people or organizations who created provided documentation and other metadata for this resource.
    associated_parties : List[Tuple[ResponsibleParty, Role]], optional
        Other people or organizations who should be associated with this resource.
    publication_date : date, optional
        The publication date of the resource.
    language : Language or str, optional
        The language in which the resource is written.
    series : str, optional
        The series from which the resource came.
    abstract : EMLTextType, optional
        A brief overview of the resource.
    keyword_set : List[EMLKeywordSet], optional
        Keyword information that describes the resource.
    additional_info: List[EMLTextType], optional
        Any extra information pertinent to the resource.
    intellectual_rights : EMLTextType, optional
        Intellectual property rights regarding usage and licensing of this resource.
    licensed : List[EMLLicense], optional
        Information identifying a well-known license for the metadata and data.
    distribution : List[EMLDistribution], optional
        Information on how the resource is distributed online and offline.
    coverage : EMLCoverage, optional
        Extent of the coverage of the resource.
    annotation : List[SemanticAnnotation], optional
        A precisely-defined semantic statement about this resource.
    """
    def __init__(
            self,
            _id: str = None, scope: Scope = Scope.DOCUMENT, system: str = None,
            referencing: bool = False, references_system: str = None,
            alternative_identifier: List[ExtensionString] = None,
            short_name: str = None, titles: List[Union[I18nString, str]] = None,
            creators: List[ResponsibleParty] = None, metadata_providers: List[ResponsibleParty] = None,
            associated_parties: List[Tuple[ResponsibleParty, Role]] = None, publication_date: dt.date = None,
            language: Union[Language, str] = None, series: str = None,
            abstract: EMLTextType = None, keyword_set: List[EMLKeywordSet] = None,
            additional_info: List[EMLTextType] = None, intellectual_rights: EMLTextType = None,
            licensed: List[EMLLicense] = None, distribution: List[EMLDistribution] = None,
            coverage: EMLCoverage = None, annotation: List[SemanticAnnotation] = None,
    ) -> None:
        super().__init__(_id, scope, system, referencing, references_system)
        if self.referencing:
            return
        self.__short_name__ = short_name
        self.__titles__: List[I18nString] = list()
        if titles is None or len(titles) == 0:
            raise ValueError("No titles provided")
        self.__titles__.extend([I18nString(title) for title in titles])
        self.__creators__: List[ResponsibleParty] = list()
        if creators is None or len(creators) == 0:
            raise ValueError("No creators provided")
        self.__creators__.extend(creators)
        self.__alternative_identifier__: List[ExtensionString] = list()
        if alternative_identifier is not None:
            self.__alternative_identifier__.extend(alternative_identifier)
        self.__metadata_provider__: List[ResponsibleParty] = list()
        if metadata_providers is not None:
            self.__metadata_provider__.extend(metadata_providers)
        self.__associated__: List[Tuple[ResponsibleParty, Role]] = list()
        if associated_parties is not None:
            self.__associated__.extend(associated_parties)
        self.__pub_date__ = publication_date
        self.__lang__ = None
        if isinstance(language, str):
            self.__lang__ = Language.get_language(language)
        elif isinstance(language, Language):
            self.__lang__ = language
        self.__series__ = series
        self.__abstract__ = abstract
        self.__keywords__: List[EMLKeywordSet] = list()
        if keyword_set is not None:
            self.__keywords__.extend(keyword_set)
        self.__add_info__: List[EMLTextType] = list()
        if additional_info is not None:
            self.__add_info__.extend(additional_info)
        self.__int_rights__ = intellectual_rights
        self.__licensed__: List[EMLLicense] = list()
        if licensed is not None:
            self.__licensed__.extend(licensed)
        self.__dist__: List[EMLDistribution] = list()
        if distribution is not None:
            self.__dist__.extend(distribution)
        self.__cover__ = coverage
        self.__annotation__: List[SemanticAnnotation] = list()
        if annotation is not None:
            self.__annotation__.extend(annotation)
        if len(self.__annotation__) > 0 and self.__id__ is None:
            raise ValueError("If annotations are given, resource must have an id.")
        return

    @property
    def title(self) -> I18nString:
        """I18nString: A brief description of the resource."""
        return self.__titles__[0]

    @property
    def titles(self) -> List[I18nString]:
        """List[I18nString]: A brief description of the resource."""
        return self.__titles__

    @property
    def creator(self) -> ResponsibleParty:
        """ResponsibleParty: The person or organization who created this resource (First listed)."""
        return self.__creators__[0]

    @property
    def creators(self) -> List[ResponsibleParty]:
        """List[ResponsibleParty]: The people or organizations who created this resource."""
        return self.__creators__

    @property
    def alternative_identifiers(self) -> List[ExtensionString]:
        """List[ExtensionString]: A secondary identifier for this entity."""
        return self.__alternative_identifier__

    @property
    def short_name(self) -> str:
        """str: A short name that describes the resource, sometimes a filename."""
        return self.__short_name__

    @property
    def metadata_provider(self) -> List[ResponsibleParty]:
        """
        List[ResponsibleParty]: The people or organizations who created provided metadata for this resource.
        """
        return self.__metadata_provider__

    @property
    def associated_party(self) -> List[Tuple[ResponsibleParty, Role]]:
        """
        List[ResponsibleParty]: Other people or organizations who should be associated with this resource.
        """
        return self.__associated__

    @property
    def publication_date(self) -> dt.date:
        """date: The publication date of the resource."""
        return self.__pub_date__

    @property
    def language(self) -> Language:
        """Language: The language in which the resource is written."""
        return self.__lang__

    @property
    def series(self) -> str:
        """str: The series from which the resource came."""
        return self.__series__

    @property
    def abstract(self) -> EMLTextType:
        """EMLTextType: A brief overview of the resource."""
        return self.__abstract__

    @property
    def keyword_set(self) -> List[EMLKeywordSet]:
        """List[EMLKeywordSet]: Keyword information that describes the resource."""
        return self.__keywords__

    @property
    def additional_info(self) -> List[EMLTextType]:
        """List[EMLTextType]: Any extra information pertinent to the resource."""
        return self.__add_info__

    @property
    def intellectual_rights(self) -> EMLTextType:
        """EMLTextType: Any extra information pertinent to the resource."""
        return self.__int_rights__

    @property
    def licensed(self) -> List[EMLLicense]:
        """List[EMLLicense]: Information identifying a well-known license for the metadata and data."""
        return self.__licensed__

    @property
    def distribution(self) -> List[EMLDistribution]:
        """List[EMLDistribution]: Information on how the resource is distributed online and offline."""
        return self.__dist__

    @property
    def coverage(self) -> EMLCoverage:
        """EMLCoverage: Extent of the coverage of the resource."""
        return self.__cover__

    @property
    def annotation(self) -> List[SemanticAnnotation]:
        """List[SemanticAnnotation]: A precisely-defined semantic statement about this resource."""
        return self.__annotation__

    def annotate(self, annotation: SemanticAnnotation) -> None:
        """
        Annotate this resource.

        Parameters
        ----------
        annotation : SemanticAnnotation
            An annotation in the :class:`eml.types.semantic_annotation.SemanticAnnotation` instance format.

        Raises
        ------
        ValueError
            If the resource does not have a unique id, it cannot be annotated.
        """
        if self.__id__ is None:
            raise ValueError("If annotations are given, resource must have an id.")
        self.__annotation__.append(annotation)
        return

    @classmethod
    def parse_kwargs(cls, element: et.Element, nmap: Dict) -> Dict:
        """
        Parse an element to generate the xml_common parameters of Resource classes.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element instance.
        nmap : Dict
            Namespace.

        Returns
        -------
        Dict
            Other arguments (kwargs) of the recourse class.
        """
        kwargs = dict()
        references_elem = element.find("references", nmap)
        kwargs["scope"] = cls.get_scope(element)
        kwargs["system"] = element.get("system", None)
        if references_elem is not None:
            kwargs["_id"] = references_elem.text
            kwargs["referencing"] = True
            kwargs["references_system"] = references_elem.get("system", None)
            return kwargs
        kwargs["_id"] = element.get("id", None)
        kwargs["referencing"] = False
        short_elem = element.find("shortName", nmap)
        if short_elem is not None:
            kwargs["short_name"] = short_elem.text
        kwargs["titles"] = list()
        for title in element.findall("title", nmap):
            kwargs["titles"].append(I18nString.parse(title, nmap))
        kwargs["creators"] = list()
        for creator in element.findall("creator", nmap):
            kwargs["creators"].append(ResponsibleParty.parse(creator, nmap))
        kwargs["alternative_identifier"] = list()
        for alt_id in element.findall("alternateIdentifier", nmap):
            kwargs["alternative_identifier"].append(ExtensionString.parse(alt_id, nmap))
        kwargs["metadata_providers"] = list()
        for meta_prov in element.findall("metadataProvider", nmap):
            kwargs["metadata_providers"].append(ResponsibleParty.parse(meta_prov, nmap))
        kwargs["associated_parties"] = list()
        for assoc_party in element.findall("associatedParty", nmap):
            kwargs["associated_parties"].append((
                ResponsibleParty.parse(assoc_party, nmap),
                ResponsibleParty.get_role(assoc_party, nmap)
            ))
        pub_date_elem = element.find("pubDate", nmap)
        if pub_date_elem is not None:
            kwargs["publication_date"] = format_datetime(pub_date_elem.text).date()
        lang_elem = element.find("language", nmap)
        if lang_elem is not None:
            kwargs["language"] = lang_elem.text
        series_elem = element.find("series", nmap)
        if series_elem is not None:
            kwargs["series"] = series_elem.text
        abs_elem = element.find("abstract", nmap)
        if abs_elem is not None:
            kwargs["abstract"] = EMLTextType.parse(abs_elem, nmap)
        kwargs["keyword_set"] = list()
        for keyword_elem in element.findall("keywordSet", nmap):
            try:
                kwargs["keyword_set"].append(EMLKeywordSet.parse(keyword_elem, nmap))
            except ValueError as e:
                warn(str(e), category=SyntaxWarning)
                pass
        kwargs["additional_info"] = list()
        for add_info in element.findall("additionalInfo", nmap):
            kwargs["additional_info"].append(EMLTextType.parse(add_info, nmap))
        int_right_elem = element.find("intellectualRights", nmap)
        if int_right_elem is not None:
            kwargs["intellectual_rights"] = EMLTextType.parse(int_right_elem, nmap)
        kwargs["licensed"] = list()
        for licensed_elem in element.findall("licensed", nmap):
            kwargs["licensed"].append(EMLLicense.parse(licensed_elem, nmap))
        kwargs["distribution"] = list()
        for dist_elem in element.findall("distribution", nmap):
            kwargs["distribution"].append(EMLDistribution.parse(dist_elem, nmap))
        cover_elem = element.find("coverage", nmap)
        if cover_elem is not None:
            try:
                kwargs["coverage"] = EMLCoverage.parse(cover_elem, nmap)
            except TypeError as e:
                warn(str(e), category=SyntaxWarning)
                pass
        kwargs["annotation"] = list()
        for annotation_elem in element.findall("annotation", nmap):
            kwargs["annotation"].append(SemanticAnnotation.parse(annotation_elem, nmap))
        return kwargs

    def to_element(self) -> et.Element:
        """
        Generate an XML Element instance from the Resource.

        Returns
        -------
        lxml.etree.Element
            XML element instance.
        """
        element = super().to_element()
        element = self._to_element_(element)
        if self.referencing:
            return element
        for alt_id in self.alternative_identifiers:
            alt_id.set_tag("alternateIdentifier")
            element.append(alt_id.to_element())
        if self.short_name is not None:
            short_elem = self.object_to_element("shortName")
            short_elem.text = self.short_name
            element.append(short_elem)
        for title in self.titles:
            title.set_tag("title")
            element.append(title.to_element())
        for creator in self.creators:
            creator.set_tag("creator")
            element.append(creator.to_element())
        for meta_prov in self.metadata_provider:
            meta_prov.set_tag("metadataProvider")
            element.append(meta_prov.to_element())
        for assoc_party in self.associated_party:
            assoc_party[0].set_tag("associatedParty")
            assoc_elem = assoc_party[0].to_element()
            role_elem = self.object_to_element("role")
            role_elem.text = assoc_party[1].to_camel_case()
            assoc_elem.append(role_elem)
            element.append(assoc_elem)
        if self.publication_date is not None:
            pub_elem = self.object_to_element("pubDate")
            pub_elem.text = self.publication_date.strftime("%Y-%m-%d")
            element.append(pub_elem)
        if self.language is not None:
            lang_elem = self.object_to_element("language")
            lang_elem.text = self.language.name.lower()
            element.append(lang_elem)
        if self.series is not None:
            series_elem = self.object_to_element("series")
            series_elem.text = self.series
            element.append(series_elem)
        if self.abstract is not None:
            self.abstract.set_tag("abstract")
            element.append(self.abstract.to_element())
        for keyword in self.keyword_set:
            element.append(keyword.to_element())
        for add_info in self.additional_info:
            add_info.set_tag("additionalInfo")
            element.append(add_info.to_element())
        if self.intellectual_rights is not None:
            self.intellectual_rights.set_tag("intellectualRights")
            element.append(self.intellectual_rights.to_element())
        for licence in self.licensed:
            element.append(licence.to_element())
        for dist in self.distribution:
            element.append(dist.to_element())
        if self.coverage is not None:
            element.append(self.coverage.to_element())
        for annotation in self.annotation:
            annotation.set_tag("annotation")
            element.append(annotation.to_element())
        return element

    def __str__(self) -> str:
        creators = "\n\t\t".join([str(creator) for creator in self.creators])
        creator_label = "Creator" if len(self.creators) <= 1 else "Creators"
        metadata_provider = "\n\t\t".join([str(metadata_provider) for metadata_provider in self.metadata_provider])
        associated_party = list()
        for assoc, role in self.associated_party:
            associated_party.append(f"\t{role}: {assoc}")
        assoc_text = "\n".join(associated_party)
        return (f"\tTitle: {self.title}\n"
                f"\t{creator_label}: {creators}\n"
                f"\tMetadataProvider: {metadata_provider}\n"
                f"{assoc_text}")



import os.path
import unittest
from typing import Tuple

from lxml import etree as et

from dwca.classes import OutsideClass
from dwca.terms import TaxonID
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestOutsideCommon(TestXML):
    @staticmethod
    def read_xml(path: str) -> Tuple[str, str]:
        with open(path, "r", encoding="utf-8") as file:
            content = file.read()
        base_file = et.fromstring(content)
        outside_xml = base_file.findall("extension", namespaces=base_file.nsmap)[0]
        text = et.tostring(outside_xml, pretty_print=True).decode("utf-8")
        return outside_xml, text

    def setUp(self) -> None:
        self.outside_xml, self.text = self.read_xml(os.path.join(PATH, os.pardir, "example_data", "meta.xml"))
        return

    def __test_parse__(self):
        outside = OutsideClass.from_string(self.text)
        self.assertEqual("speciesprofile.txt", outside.filename, "Files parse incorrectly")
        self.assertEqual(0, outside.id, "Files parse incorrectly")
        self.assertEqual("UTF-8", outside.__encoding__, "Encoding parse incorrectly")
        self.assertEqual("\n", outside.__lines_end__, "Lines end parse incorrectly")
        self.assertEqual("\t", outside.__fields_end__, "Field end parse incorrectly")
        self.assertEqual("", outside.__fields_enclosed__, "Field enclosed parse incorrectly")
        self.assertEqual(1, outside.__ignore_header_lines__, "Ignore header parse incorrectly")
        self.assertEqual("http://rs.gbif.org/terms/1.0/SpeciesProfile", outside.uri, "Row type parse incorrectly")
        self.assertEqual(5, len(outside.__fields__), "Fields parse incorrectly")
        self.assertEqualTree(self.outside_xml, outside.to_element(), "Error on element conversion")

    def __test_none__(self):
        self.assertIsNone(OutsideClass.parse(None, {}), "Object parsed from nothing")

    def __test_merge__(self):
        xml_file_1, text_1 = self.read_xml(os.path.join(PATH, os.pardir, "example_data", "dataset1_meta.xml"))
        xml_file_2, text_2 = self.read_xml(os.path.join(PATH, os.pardir, "example_data", "dataset2_meta.xml"))
        extension1 = OutsideClass.from_string(text_1)
        with open(
                os.path.join(PATH, os.pardir, "example_data", "extension_example1.txt"),
                "r", encoding="utf-8"
        ) as file:
            extension1.read_file(file.read())
        extension2 = OutsideClass.from_string(text_2)
        with open(
                os.path.join(PATH, os.pardir, "example_data", "extension_example2.txt"),
                "r", encoding="utf-8"
        ) as file:
            extension2.read_file(file.read())
        merged = OutsideClass.merge(extension1, extension2)
        self.assertEqual(len(merged), len(extension1) + len(extension2), "Incorrect merged of taxa files.")
        self.assertCountEqual(
            merged.fields,
            extension1.fields,
            "Incorrect number of fields"
        )

    def __test_set_primary_key__(self):
        outside_class = OutsideClass.from_string(self.text)
        self.assertIsNone(outside_class.__primary_key__, "Primary key from nowhere.")
        outside_class.set_primary_key("Taxon")
        self.assertIsNotNone(outside_class.__primary_key__, "Primary key not set.")
        self.assertEqual("Taxon", outside_class.__primary_key__, "Incorrect Primary key.")

    def __test_sql_table__(self):
        self.outside_xml, self.text = self.read_xml(os.path.join(PATH, os.pardir, "example_data", "meta.xml"))
        outside_class = OutsideClass.from_string(self.text)
        with self.assertRaisesRegex(
            RuntimeError,
            "Extension",
        ):
            _ = outside_class.sql_table
        outside_class.set_primary_key("Taxon")
        with self.assertRaisesRegex(
            RuntimeError,
            "Extension",
        ):
            _ = outside_class.sql_table
        outside_class.set_core_field(TaxonID(5))
        self.assertEqual(self.normalize_sql("""
        CREATE TABLE "SpeciesProfile" (
            "taxonID" VARCHAR,
            "isMarine" VARCHAR,
            "isFreshwater" VARCHAR,
            "isTerrestrial" VARCHAR,
            "isExtinct" VARCHAR,
            PRIMARY KEY ("taxonID"),
            FOREIGN KEY ("taxonID") REFERENCES "Taxon"
        );
        """), self.normalize_sql(outside_class.sql_table), "Wrong SQL table.")

    def __test_insert__(self):
        text_1 = """
        <extension encoding="UTF-8" linesTerminatedBy="\\n" fieldsTerminatedBy="\\t" fieldsEnclosedBy="" ignoreHeaderLines="1" rowType="http://example.org/terms/extension1">
        <files>
            <location>extension1.txt</location>
        </files>
        <coreid index="0" /><field index="1" term="http://example.org/terms/column1"/>
        <field index="2" term="http://example.org/terms/column2"/>
        </extension>
        """
        extension1 = OutsideClass.from_string(text_1)
        with self.assertRaisesRegex(
            RuntimeError,
            "Extension",
        ):
            next(extension1.insert_sql)
        extension1.set_core_field(TaxonID(5))
        with open(
                os.path.join(PATH, os.pardir, "example_data", "extension_example1.txt"),
                "r", encoding="utf-8"
        ) as file:
            extension1.read_file(file.read())
        text_2 = """
        <extension encoding="UTF-8" linesTerminatedBy="\\n" fieldsTerminatedBy="\\t" fieldsEnclosedBy="" ignoreHeaderLines="1" rowType="http://example.org/terms/extension2">
        <files>
          <location>extension2.txt</location>
        </files>
        <coreid index="0" /><field index="1" term="http://example.org/terms/column1"/>
        </extension>
        """
        extension2 = OutsideClass.from_string(text_2)
        extension2.set_core_field(TaxonID(5))
        with open(
                os.path.join(PATH, os.pardir, "example_data", "extension_example2.txt"),
                "r", encoding="utf-8"
        ) as file:
            extension2.read_file(file.read())
        expected_values_1 = [
            ("urn:lsid:example.org:taxname:0", "true", "false"),
            ("urn:lsid:example.org:taxname:8", "false", "true"),
            ("urn:lsid:example.org:taxname:19", "true", "true")
        ]
        self.assertEqual(3, len(list(extension1.insert_sql)), "Not the right amount of rows.")
        for (statement, values), expected in zip(extension1.insert_sql, expected_values_1):
            self.assertEqual(
                self.normalize_sql("""
                INSERT INTO "extension1" (
                    "taxonID",
                    "column1",
                    "column2"
                ) VALUES (%s, %s, %s)
                """), self.normalize_sql(statement), "Statement do not match"
            )
            self.assertEqual(expected, values, "Values do not match")
        expected_values_2 = [
            ("urn:lsid:example.org:taxname:20", "true"),
            ("urn:lsid:example.org:taxname:27", "true"),
            ("urn:lsid:example.org:taxname:41", "false")
        ]
        self.assertEqual(3, len(list(extension2.insert_sql)), "Not the right amount of rows.")
        for (statement, values), expected in zip(extension2.insert_sql, expected_values_2):
            self.assertEqual(
                self.normalize_sql("""
                INSERT INTO "extension2" (
                    "taxonID",
                    "column1"
                ) VALUES (%s, %s)
                """), self.normalize_sql(statement), "Statement do not match"
            )
            self.assertEqual(expected, values, "Values do not match")


if __name__ == '__main__':
    unittest.main()



import unittest

from lxml import etree as et

from eml.resources.distribution import EMLOffline
from test_xml.test_xml import TestXML


class TestOffline(TestXML):
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
    }

    def test_parse_none(self):
        self.assertIsNone(EMLOffline.parse(None, {}), "EMLOffline from nowhere")

    def test_parse_invalid(self):
        text_xml = """
<offline>
    <mediumDensity>High Density (HD)</mediumDensity>
    <mediumDensityUnits>B/cm</mediumDensityUnits>
    <mediumVolume>650 MB</mediumVolume>
    <mediumFormat>NTFS</mediumFormat>
    <mediumNote>any additional pertinent information about the media</mediumNote>
</offline>
        """
        self.assertRaises(
            AttributeError,
            EMLOffline.from_string,
            text_xml
        )

    def test_parse_minimum(self):
        text_xml = """
<offline>
    <mediumName>Tape, 3.5 inch Floppy Disk, hardcopy</mediumName>
</offline>
        """
        offline = EMLOffline.from_string(text_xml)
        self.assertEqual("Tape, 3.5 inch Floppy Disk, hardcopy", offline.medium_name,
                         "Incorrect parsed of medium name")
        self.assertIsNone(offline.medium_density, "Incorrect parsed of medium density")
        self.assertIsNone(offline.medium_density_units, "Incorrect parsed of medium density units")
        self.assertIsNone(offline.medium_volume, "Incorrect parsed of medium volume")
        self.assertEqual(0, len(offline.medium_format), "Incorrect parsed of medium format")
        self.assertIsNone(offline.medium_note, "Incorrect parsed of medium note")
        self.assertEqualTree(et.fromstring(text_xml), offline.to_element(), "Error on to element")

    def test_parse_complete(self):
        text_xml = """
<offline>
    <mediumName>Tape, 3.5 inch Floppy Disk, hardcopy</mediumName>
    <mediumDensity>High Density (HD)</mediumDensity>
    <mediumDensityUnits>B/cm</mediumDensityUnits>
    <mediumVolume>650 MB</mediumVolume>
    <mediumFormat>NTFS</mediumFormat>
    <mediumNote>any additional pertinent information about the media</mediumNote>
</offline>
        """
        offline = EMLOffline.from_string(text_xml)
        self.assertEqual("Tape, 3.5 inch Floppy Disk, hardcopy", offline.medium_name,
                         "Incorrect parsed of medium name")
        self.assertEqual("High Density (HD)", offline.medium_density, "Incorrect parsed of medium density")
        self.assertEqual("B/cm", offline.medium_density_units, "Incorrect parsed of medium density units")
        self.assertEqual("650 MB", offline.medium_volume, "Incorrect parsed of medium volume")
        self.assertEqual(1, len(offline.medium_format), "Incorrect parsed of medium format")
        self.assertEqual("NTFS", offline.medium_format[0], "Incorrect parsed of medium format")
        self.assertEqual(
            "any additional pertinent information about the media",
            offline.medium_note,
            "Incorrect parsed of medium note"
        )
        self.assertEqualTree(et.fromstring(text_xml), offline.to_element(), "Error on to element")

    def test_parse_more_formats(self):
        text_xml = """
<offline>
    <mediumName>Tape, 3.5 inch Floppy Disk, hardcopy</mediumName>
    <mediumDensity>High Density (HD)</mediumDensity>
    <mediumDensityUnits>B/cm</mediumDensityUnits>
    <mediumVolume>650 MB</mediumVolume>
    <mediumFormat>NTFS</mediumFormat>
    <mediumFormat>FAT32</mediumFormat>
    <mediumFormat>EXT2</mediumFormat>
    <mediumFormat>QIK80</mediumFormat>
    <mediumNote>any additional pertinent information about the media</mediumNote>
</offline>
        """
        offline = EMLOffline.from_string(text_xml)
        self.assertEqual("Tape, 3.5 inch Floppy Disk, hardcopy", offline.medium_name,
                         "Incorrect parsed of medium name")
        self.assertEqual("High Density (HD)", offline.medium_density, "Incorrect parsed of medium density")
        self.assertEqual("B/cm", offline.medium_density_units, "Incorrect parsed of medium density units")
        self.assertEqual("650 MB", offline.medium_volume, "Incorrect parsed of medium volume")
        self.assertEqual(4, len(offline.medium_format), "Incorrect parsed of medium format")
        self.assertEqual(
            "NTFS, FAT32, EXT2, QIK80".split(", "),
            offline.medium_format,
            "Incorrect parsed of medium format"
        )
        self.assertEqual(
            "any additional pertinent information about the media",
            offline.medium_note,
            "Incorrect parsed of medium note"
        )
        self.assertEqualTree(et.fromstring(text_xml), offline.to_element(), "Error on to element")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Dict, List, Union
from warnings import warn

from lxml import etree as et

from xml_common.utils import CamelCaseEnum
from eml.types import EMLObject, _NoTagObject, Scope, IndividualName, OrganizationName, PositionName, EMLAddress, \
    EMLPhone, I18nString


class Role(CamelCaseEnum):
    """
    The role the party played with respect to the resource.
    """
    AUTHOR = 0
    CONTENT_PROVIDER = 1
    CUSTODIAN_STEWARD = 2
    DISTRIBUTOR = 3
    EDITOR = 4
    METADATA_PROVIDER = 5
    ORIGINATOR = 6
    POINT_OF_CONTACT = 7
    PRINCIPAL_INVESTIGATOR = 8
    PROCESSOR = 9
    PUBLISHER = 10
    USER = 11


class ResponsibleParty(EMLObject, _NoTagObject):
    """
    The individual, organization, or role associated with a resource.

    Parameters
    ----------
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    individual_name : IndividualName, optional
        The full name of the person being described.
    organization_name : OrganizationName, optional
        The full name of the organization being described.
    position_name : PositionName, optional
        The name of the title or position associated with the resource.
    address : List[EMLAddress], optional
        A list of full addresses information for a given responsible party entry.
    phone : List[EMLPhone], optional
        A list of phone information about the contact.
    mail : List[str or I18nString], optional
        A list of email addresses of the contact.
    url : List[str], optional
        A list of links to associated online information, usually a website.
    """
    def __init__(
            self, _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
            referencing: bool = False,
            references_system: str = None,
            individual_name: IndividualName = None,
            organization_name: OrganizationName = None,
            position_name: PositionName = None,
            address: List[EMLAddress] = None,
            phone: List[EMLPhone] = None,
            mail: List[Union[str, I18nString]] = None,
            url: List[str] = None,
    ) -> None:
        EMLObject.__init__(self, _id, scope, system, referencing, references_system)
        _NoTagObject.__init__(self)
        if not self.referencing:
            assert_msg = ("At least one of the following must be given: individual_name, "
                          "organization_name or position_name")
            assert individual_name is not None or organization_name is not None or position_name is not None, assert_msg
        self.__individual__ = individual_name
        self.__organization__ = organization_name
        self.__position__ = position_name
        self.__address__: List[EMLAddress] = list()
        if address is not None:
            self.__address__.extend(address)
        self.__phone__: List[EMLPhone] = list()
        if phone is not None:
            self.__phone__.extend(phone)
        self.__mail__: List[I18nString] = list()
        if mail is not None:
            self.__mail__.extend([I18nString(m) for m in mail])
        self.__url__: List[str] = list()
        if url is not None:
            self.__url__.extend(url)
        return

    @property
    def individual_name(self) -> IndividualName:
        """IndividualName: the full name of the person being described."""
        return self.__individual__

    @property
    def organization_name(self) -> OrganizationName:
        """OrganizationName: The full name of the organization being described."""
        return self.__organization__

    @property
    def position_name(self) -> PositionName:
        """PositionName: The name of the title or position associated with the resource."""
        return self.__position__

    @property
    def address(self) -> List[EMLAddress]:
        """List[EMLAddress]: A list of full addresses information for a given responsible party entry."""
        return self.__address__

    @property
    def phone(self) -> List[EMLPhone]:
        """List[EMLPhone]: A list of phon information about the contact."""
        return self.__phone__

    @property
    def mail(self) -> List[I18nString]:
        """List[I18nString]: A list of email addresses of the contact."""
        return self.__mail__

    @property
    def url(self) -> List[str]:
        """List[str]: A list of links to associated online information, usually a website."""
        return self.__url__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> ResponsibleParty:
        """
        Generate a Responsible Party referencing another Responsible Party.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        ResponsibleParty
            Object parsed that reference another responsible party on document.
        """
        references = element.find("references")
        return ResponsibleParty(
            _id=references.text,
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None)
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> ResponsibleParty:
        """
        Generate a Responsible Party that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        ResponsibleParty
            Object parsed.
        """
        return ResponsibleParty(
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=False,
            individual_name=IndividualName.parse(element.find("individualName", None), nmap),
            organization_name=OrganizationName.parse(element.find("organizationName", None), nmap),
            position_name=PositionName.parse(element.find("positionName", None), nmap),
            address=[EMLAddress.parse(address, nmap) for address in element.findall("address", nmap)],
            phone=[EMLPhone.parse(phone, nmap) for phone in element.findall("phone", nmap)],
            mail=[I18nString.parse(mail, nmap) for mail in element.findall("electronicMailAddress", nmap)],
            url=[url.text for url in element.findall("onlineUrl", nmap)],
        )

    def to_element(self) -> et.Element:
        """
        Generate a lxml.tree.Element from the Responsible Party object.

        Returns
        -------
        lxml.tree.Element
            Object in the XML Element format.

        Raises
        ------
        RuntimeError
            If tag is not set before calling this method.
        """
        resp_party = _NoTagObject.to_element(self)
        resp_party = self._to_element_(resp_party)
        references = self.generate_references_element()
        if references is None:
            if self.individual_name is not None:
                resp_party.append(self.individual_name.to_element())
            if self.organization_name is not None:
                resp_party.append(self.organization_name.to_element())
            if self.position_name is not None:
                resp_party.append(self.position_name.to_element())
            for address in self.address:
                resp_party.append(address.to_element())
            for phone in self.phone:
                resp_party.append(phone.to_element())
            for email in self.mail:
                email.set_tag("electronicMailAddress")
                resp_party.append(email.to_element())
            for url in self.__url__:
                url_elem = et.Element("onlineUrl")
                url_elem.text = url
                resp_party.append(url_elem)
        return resp_party

    @classmethod
    def get_role(cls, element: et.Element, nmap: Dict) -> Role:
        """
        Get role for an XML element.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element with role on it.
        nmap : Dict
            Namespace.

        Returns
        -------
        Role
            Value of the role inside the XML element.
        """
        role_elem = element.find("role", nmap)
        try:
            return Role.get_enum(role_elem.text)
        except ValueError:
            link = "https://eml.ecoinformatics.org/schema/eml-party_xsd.html#RoleType"
            warn(f"Role {role_elem.text} is not supported, "
                 f"check {link} for the full list of choices. "
                 f"Using contentProvider as default.", category=SyntaxWarning)
            return Role.CONTENT_PROVIDER

    def __str__(self) -> str:
        position_name = "" if self.position_name is None else f"{self.position_name}"
        if self.organization_name is None:
            organization_name = position_name
        else:
            if len(position_name) == 0:
                organization_name = f"{self.organization_name}"
            else:
                organization_name = f"{position_name} at {self.organization_name}"
        if self.individual_name is None:
            return organization_name
        else:
            if len(organization_name) > 0:
                organization_name = f" ({organization_name})"
            return f"{self.individual_name}{organization_name}"



import os
import sys
import unittest
from unittest.mock import patch

from dwca.base import DarwinCoreArchive
from test_dwca.test_dwca_common import TestDWCACommon

PATH = os.path.abspath(os.path.dirname(__file__))

orig_import = __import__


def import_mock(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'pandas':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


class TestDWCANoPandas(TestDWCACommon):
    @patch('builtins.__import__', side_effect=import_mock)
    def test_minimal(self, mock_import):
        super().__test_minimal__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_attributes(self, mock_import):
        super().__test_attributes__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_merge(self, mock_import):
        super().__test_merge__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_set_core(self, mock_import):
        core = self.object.core
        self.assertEqual(163460, len(core), "Wrong number of entries")
        core.__entries__ = core.__entries__[0:100]
        self.assertEqual(100, len(core), "Core has not been shorted to 100 entries")
        self.object.core = core
        self.assertEqual(100, len(self.object.core), "Core has not been shorted on DwC-A object")
        for extension in self.object.extensions:
            self.assertGreaterEqual(100, len(extension), f"Extension {extension.uri} has not been shorted")


if __name__ == '__main__':
    unittest.main()



def read_string(content: str | bytes, encoding: str = "utf-8") -> str:
    """
    Read a string or a byte into string.

    Parameters
    ----------
    content : str | bytes
        Content as a string or byte.
    encoding : str, optional
        Encoding of the content.

    Returns
    -------
    str
        Content as string.
    """
    try:
        return content.decode()
    except ValueError:
        return content



from __future__ import annotations
from typing import Dict, List, Union

from lxml import etree as et

from eml.types import ProcedureStep
from xml_common import XMLObject


class Methods(XMLObject):
    """
    Class representing the methods field documents scientific methods used in the collection of this dataset.

    Parameters
    ----------
    method_steps : List[ProcedureStep]
        A list of description of a specific step of the method. At least one must be given.
    data_sources : List[List[EMLDataset] | None], optional
        The sources of data used by each method step.

    Raises
    ------
    RuntimeError
        At least one step must be given.
    """
    class MethodStep(ProcedureStep):
        """
        Represent the description of a specific step of the method.

        Parameters
        ----------
        procedure_step : ProcedureStep
            A description of a specific step.
        data_source : List[EMLDataset], optional
            Sources of data used by this MethodStep.
        """
        PRINCIPAL_TAG = "methodStep"
        def __init__(
                self,
                procedure_step: ProcedureStep,
                data_source: List[EMLDataset] = None
        ) -> None:
            super().__init__(
                description=procedure_step.description,
                citation=procedure_step.citation,
                protocol=procedure_step.protocol,
                instrumentation=procedure_step.instrumentation,
                software=procedure_step.software,
                sub_step=procedure_step.sub_step
            )
            self.__data_source__ = list()
            if data_source is not None:
                self.__data_source__.extend(data_source)
            return

        @property
        def data_source(self) -> List[EMLDataset]:
            """List[EMLDataset]: Sources of data used by this MethodStep."""
            return self.__data_source__

        @classmethod
        def parse(cls, element: et.Element, nmap: Dict) -> Methods.MethodStep | None:
            """
            Parses a lxml element into a Method Step object.

            Parameters
            ----------
            element : lxml.tree.Element
                Object in the Element format to be parsed into a Python object.
            nmap : Dict
                Dictionary of namespace.

            Returns
            -------
            Methods.MethodStep
                Object parsed.
            """
            procedure_step = super().parse(element, nmap)
            data_source = list()
            data_elem = element.findall("dataSource", nmap)
            if len(data_elem) > 0:
                from eml.resources import EMLDataset
                for ds in data_elem:
                    data_source.append(EMLDataset.parse(ds, nmap))
            return Methods.MethodStep(
                procedure_step=procedure_step,
                data_source=data_source
            )

        def to_element(self) -> et.Element:
            """
            Generates a lxml.tree.Element from the Method Step object.

            Returns
            -------
            lxml.tree.Element
                Methods instance in the Element format.
            """
            self.set_tag(self.PRINCIPAL_TAG)
            element = super().to_element()
            for ds in self.data_source:
                ds.PRINCIPAL_TAG = "dataSource"
                element.append(ds.to_element())
            return element

    PRINCIPAL_TAG = "methods"

    def __init__(
            self, method_steps: List[ProcedureStep],
            data_sources: List[Union[List[EMLDataset], None]] = None
    ) -> None:
        super().__init__()
        self.__m_steps__ = list()
        for i, method_step in enumerate(method_steps):
            ms_kwargs = {"procedure_step": method_step}
            if data_sources is not None:
                try:
                    ms_kwargs["data_source"] = data_sources[i]
                except IndexError:
                    pass
            self.__m_steps__.append(Methods.MethodStep(**ms_kwargs))
        if len(self.__m_steps__) == 0:
            raise TypeError("At least one step must be given.")
        return

    @property
    def method_steps(self) -> List[Methods.MethodStep]:
        """List[MethodStep]: A list of description of a specific step of the method."""
        return self.__m_steps__

    @property
    def sampling(self) -> None:
        """None: Description of sampling procedures including the geographic, temporal and taxonomic coverage of the study."""
        # TODO: implement Sampling (https://eml.ecoinformatics.org/schema/eml-methods_xsd.html#MethodsType_sampling)
        return

    @property
    def quality_control(self) -> List[ProcedureStep]:
        """ProcedureStep: Information on possible errors or on the quality of a data set."""
        # TODO: implement Quality Control (https://eml.ecoinformatics.org/schema/eml-methods_xsd.html#MethodsType_qualityControl)
        return list()

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> Methods | None:
        """
        Parses a lxml element into a Methods object.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Python object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        Methods
            Object parsed.
        """
        if element is None:
            return None
        method_steps = list()
        data_sources = list()
        for ms in element.findall("methodStep", nmap):
            method_steps.append(Methods.MethodStep.parse(ms, nmap))
            ds_elements = ms.findall("dataSource", nmap)
            ds = list()
            if len(ds_elements) > 0:
                from eml.resources import EMLDataset
                for ds_elem in ds_elements:
                    ds.append(EMLDataset.parse(ds_elem, nmap))
            data_sources.append(ds)
        return Methods(
            method_steps=method_steps,
            data_sources=data_sources
        )

    def to_element(self) -> et.Element:
        """
        Generates a lxml.tree.Element from the Methods object.

        Returns
        -------
        lxml.tree.Element
            Methods instance in the Element format.
        """
        element = super().to_element()
        for method_step in self.method_steps:
            element.append(method_step.to_element())
        return element



import unittest

from lxml import etree as et

from dwca.terms import MeasurementValue
from test_xml.test_xml import TestXML


class TestAnyField(TestXML):
    def test_parse(self):
        measurement_value_xml = """<field index="3" term="http://rs.tdwg.org/dwc/terms/measurementValue"/>"""
        measurement_value = MeasurementValue.from_string(measurement_value_xml)
        self.assertEqual(
            "http://rs.tdwg.org/dwc/terms/measurementValue",
            measurement_value.uri,
            "Wrong URI"
        )
        self.assertEqual(3, measurement_value.index, "Wrong index")
        self.assertEqualTree(et.fromstring(measurement_value_xml), measurement_value.to_element(), "Wrong generated element.")
        pass

    def test_format(self):
        measurement_value = MeasurementValue(3)
        value = "1"
        with self.assertWarnsRegex(Warning, "<Any>"):
            self.assertEqual(
                value, measurement_value.format(value), "Error formatting Any"
            )
        value = "14.5"
        with self.assertWarnsRegex(Warning, "<Any>"):
            self.assertEqual(
                value, measurement_value.format(value), "Error formatting Any"
            )
        value = "UV-light"
        with self.assertWarnsRegex(Warning, "<Any>"):
            self.assertEqual(value, measurement_value.format(value), "Error formatting Any")


if __name__ == '__main__':
    unittest.main()



import datetime as dt
from typing import List

from dwca.terms import Field


class ChronometricAgeID(Field):
    """
    An identifier for the set of information associated with a ChronometricAge.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimChronometricAge(Field):
    """
    The verbatim age for a specimen, whether reported by a dating assay, associated references, or legacy information.

    For example, this could be the radiocarbon age
    as given in an AMS dating report. This could
    also be simply what is reported as the age of
    a specimen in legacy collections data.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/verbatimChronometricAge"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ChronometricAgeProtocol(Field):
    """
    A description of or reference to the methods used to determine the chronometric age.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeProtocol"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class UncalibratedChronometricAge(Field):
    """
    The output of a dating assay before it is calibrated into an age using a specific conversion protocol.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/uncalibratedChronometricAge"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ChronometricAgeConversionProtocol(Field):
    """
    The method used for converting the UncalibratedChronometricAge into a chronometric age in years.

    As captured in the ``EarliestChronometricAge``, ``EarliestChronometricAgeReferenceSystem``,
    ``LatestChronometricAge``, and ``LatestChronometricAgeReferenceSystem`` fields.
    For example, calibration of conventional radiocarbon age or the
    currently accepted age range of a cultural or geological period.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeConversionProtocol"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EarliestChronometricAge(Field):
    """
    The maximum/earliest/oldest possible age of a specimen as determined by a dating method.

    The expected unit for this field is years. This field, if populated,
    must have an associated ``EarliestChronometricAgeReferenceSystem``.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/earliestChronometricAge"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EarliestChronometricAgeReferenceSystem(Field):
    """
    The reference system associated with the EarliestChronometricAge.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/chrono/terms/earliestChronometricAgeReferenceSystem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class LatestChronometricAge(Field):
    """
    The minimum/latest/youngest possible age of a specimen as determined by a dating method.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/latestChronometricAge"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LatestChronometricAgeReferenceSystem(Field):
    """
    The reference system associated with the LatestChronometricAge.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/chrono/terms/latestChronometricAgeReferenceSystem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class ChronometricAgeUncertaintyInYears(Field):
    """
    The temporal uncertainty of the EarliestChronometricAge and LatestChronometicAge in years.

    The expected unit for this field is years. The value
    in this field is number of years before and after the
    values given in the earliest and latest chronometric
    age fields within which the actual values are estimated to be.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeUncertaintyInYears"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ChronometricAgeUncertaintyMethod(Field):
    """
    The method used to generate the value of ChronometricAgeUncertaintyInYears.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeUncertaintyMethod"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MaterialDated(Field):
    """
    A description of the material on which the chronometricAgeProtocol was actually performed, if known.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/materialDated"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MaterialDatedID(Field):
    """
    An identifier for the MaterialSample on which the chronometricAgeProtocol was performed, if applicable.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/materialDatedID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MaterialDatedRelationship(Field):
    """
    The relationship of the MaterialDated to the subject of the ChronometricAge record.

    From which the ChronometricAge of the subject is inferred.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/chrono/terms/materialDatedRelationship"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class ChronometricAgeDeterminedBy(Field):
    """
    A list of names of people, groups, or organizations who determined the ChronometricAge.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeDeterminedBy"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ChronometricAgeDeterminedDate(Field):
    """
    The date on which the ChronometricAge was determined.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeDeterminedDate"
    TYPE = dt.datetime

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ChronometricAgeReferences(Field):
    """
    A list of identifiers of literature associated with the ChronometricAge.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeReferences"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ChronometricAgeRemarks(Field):
    """
    Notes or comments about the ChronometricAge.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: Any, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/chrono/terms/chronometricAgeRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



import datetime as dt
import unittest

import lxml.etree as et

from eml.types import Maintenance, MaintUpFreqType
from test_xml.test_xml import TestXML


class TestMaintenance(TestXML):
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
    }

    def test_parse_none(self):
        self.assertIsNone(Maintenance.parse(None, {}), "Change History from None.")

    def test_parse_simple(self):
        text_xml = """
<maintenance>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.</para>
    </description>
</maintenance>
        """
        maintenance = Maintenance.from_string(text_xml)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.",
            maintenance.description.paragraphs[0],
            "Error on parsing description."
        )
        self.assertIsNone(maintenance.maintenance_update_frequency, "Frequency from nowhere.")
        self.assertEqual(0, len(maintenance.change_history), "Change history from nowhere.")
        self.assertEqualTree(et.fromstring(text_xml), maintenance.to_element(), "Error on to element.")

    def test_parse_complete(self):
        text_xml = """
<maintenance>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.</para>
    </description>
    <maintenanceUpdateFrequency>annually</maintenanceUpdateFrequency>
    <changeHistory>
        <changeScope>Scope</changeScope>
        <oldValue>Document</oldValue>
        <changeDate>2001</changeDate>
    </changeHistory>
</maintenance>
        """
        maintenance = Maintenance.from_string(text_xml)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.",
            maintenance.description.paragraphs[0],
            "Error on parsing description."
        )
        self.assertEqual(MaintUpFreqType.ANNUALLY, maintenance.maintenance_update_frequency, "Error on parsing frequency.")
        self.assertEqual(1, len(maintenance.change_history), "Error on parsing change history.")
        self.assertEqual("Scope", maintenance.change_history[0].change_scope, "Error on parsing change history (\"change\").")
        self.assertEqualTree(et.fromstring(text_xml), maintenance.to_element(), "Error on to element.")

    def test_parse_multiple_change_history(self):
        text_xml = """
<maintenance>
    <description>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.</para>
    </description>
    <maintenanceUpdateFrequency>annually</maintenanceUpdateFrequency>
    <changeHistory>
        <changeScope>Scope</changeScope>
        <oldValue>Document</oldValue>
        <changeDate>2001</changeDate>
    </changeHistory>
    <changeHistory>
        <changeScope>Scope</changeScope>
        <oldValue>System</oldValue>
        <changeDate>2002</changeDate>
    </changeHistory>
    <changeHistory>
        <changeScope>Scope</changeScope>
        <oldValue>Document</oldValue>
        <changeDate>2003</changeDate>
    </changeHistory>
</maintenance>
        """
        maintenance = Maintenance.from_string(text_xml)
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.",
            maintenance.description.paragraphs[0],
            "Error on parsing description."
        )
        self.assertEqual(MaintUpFreqType.ANNUALLY, maintenance.maintenance_update_frequency,
                         "Error on parsing frequency.")
        self.assertEqual(3, len(maintenance.change_history), "Error on parsing change history.")
        self.assertEqual("Scope", maintenance.change_history[0].change_scope,
                         "Error on parsing change history (\"change\").")
        self.assertEqual("Document", maintenance.change_history[0].old_value,
                         "Error on parsing change history (\"old value\", 1).")
        self.assertEqual(dt.date(year=2001, month=1, day=1), maintenance.change_history[0].change_date,
                         "Error on parsing change history (\"date\", 1).")
        self.assertEqual("System", maintenance.change_history[1].old_value,
                         "Error on parsing change history (\"old value\", 2).")
        self.assertEqual(dt.date(year=2002, month=1, day=1), maintenance.change_history[1].change_date,
                         "Error on parsing change history (\"date\", 2).")
        self.assertEqual("Document", maintenance.change_history[2].old_value,
                         "Error on parsing change history (\"old value\", 3).")
        self.assertEqual(dt.date(year=2003, month=1, day=1), maintenance.change_history[2].change_date,
                         "Error on parsing change history (\"date\", 3).")
        self.assertEqualTree(et.fromstring(text_xml), maintenance.to_element(), "Error on to element.")


if __name__ == '__main__':
    unittest.main()



import unittest

from lxml import etree as et

from eml.resources import EMLCitation
from test_xml.test_xml import TestXML


class TestEMLCitation(TestXML):
    DEFAULT_TAGS = {
        "scope": "document"
    }

    def test_parse_referencing(self):
        text_xml = """
<citation>
    <references>cite-id</references>
</citation>
        """
        citation = EMLCitation.from_string(text_xml)
        self.assertTrue(citation.referencing, "Referencing parse incorrectly")
        self.assertEqual("cite-id", citation.id, "Id parse incorrectly")
        self.assertEqualTree(et.fromstring(text_xml), citation.to_element(), "Error on to element")


if __name__ == '__main__':
    unittest.main()



import unittest

from lxml import etree as et

from dwca.terms import IndividualCount
from test_xml.test_xml import TestXML


class TestIntField(TestXML):
    def test_parse(self):
        ind_count_xml = """<field index="3" term="http://rs.tdwg.org/dwc/terms/individualCount"/>"""
        ind_count = IndividualCount.from_string(ind_count_xml)
        self.assertEqual(
            "http://rs.tdwg.org/dwc/terms/individualCount",
            ind_count.uri,
            "Wrong URI"
        )
        self.assertEqual(3, ind_count.index, "Wrong index")
        self.assertEqualTree(et.fromstring(ind_count_xml), ind_count.to_element(), "Wrong generated element.")
        pass

    def test_format(self):
        ind_count = IndividualCount(3)
        value = "30"
        self.assertEqual(
            30, ind_count.format(value), "Error formatting integer"
        )
        value = "Invalid integer"
        self.assertIsNone(ind_count.format(value), "Error formatting invalid integer")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import Dict

from lxml import etree as et

from xml_common import XMLObject


class EMLInline(XMLObject):
    """
    Inline distribution.
    """
    PRINCIPAL_TAG = "inline"

    def __init__(self) -> None:
        super().__init__()
        return

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLInline | None:
        """
        Generate an EMLInline instance from an XML element.

        Parameters
        ----------
        element : lxml.etree.Element
            XML Element to be parsed.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLInline
            An instance with the information of the XML element.
        """
        if element is None:
            return None
        else:
            return EMLInline()

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance with the information of the EMLInline object.

        Returns
        -------
        lxml.etree.Element
            XML element instance.
        """
        return super().to_element()



import unittest
from lxml import etree as et

from xml_common.utils import Language
from eml.types import EMLSection
from test_xml.test_xml import TestXML


class TestSection(TestXML):
    PARAGRAPH = ("Lorem ipsum dolor sit amet, consectetur "
                 "adipiscing elit. Sed euismod urna sit amet "
                 "pulvinar porttitor. Ut aliquet lectus sit amet.")
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
    }

    def test_parse_simple(self):
        text_xml = f"""
<section>
    <title>A simple section</title>
    <para>{self.PARAGRAPH}</para>
</section>
        """
        section = EMLSection.from_string(text_xml)
        self.assertEqual("A simple section", section.title, "Title parsed incorrectly")
        self.assertEqual(Language.ENG, section.title.language, "Title language parsed incorrectly")
        self.assertEqual(self.PARAGRAPH, section.paragraph, "Paragraph parsed incorrectly")
        self.assertIsNone(section.section, "Section generated from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), section.to_element(), "To element incorrect")

    def test_parse_empty(self):
        text_xml = "<section></section>"
        self.assertRaises(ValueError, EMLSection.from_string, text_xml)

    def test_parse_none(self):
        self.assertIsNone(EMLSection.parse(None, {}), "Error on parse None")

    def test_equal(self):
        section = EMLSection(
            title="A simple section",
            paragraph="A simple paragraph"
        )
        section_esp = EMLSection(
            title="A simple section",
            paragraph="A simple paragraph",
            language=Language.ESP
        )
        self.assertEqual(section, section_esp, "Compare language")
        self.assertNotEqual(3, section_esp, "Error on equal implementation")

    def test_parse_no_title(self):
        text_xml = f"""
<section>
    <para>{self.PARAGRAPH}</para>
</section>
        """
        section = EMLSection.from_string(text_xml)
        self.assertIsNone(section.title, "Title generated from nowhere")
        self.assertEqual(self.PARAGRAPH, section.paragraph, "Paragraph parsed incorrectly")
        self.assertIsNone(section.section, "Section generated from nowhere")
        self.assertEqualTree(et.fromstring(text_xml), section.to_element(), "To element incorrect")

    def test_parse_complete(self):
        text_xml = f"""
<section>
    <title>A simple section</title>
    <section>
        <para>{self.PARAGRAPH}</para>
    </section>
</section>
        """
        section = EMLSection.from_string(text_xml)
        self.assertEqual("A simple section", section.title, "Title parsed incorrectly")
        self.assertEqual(Language.ENG, section.title.language, "Title language parsed incorrectly")
        self.assertEqual(EMLSection(paragraph=self.PARAGRAPH), section.section, "Error on equal implementation")
        self.assertEqualTree(et.fromstring(text_xml), section.to_element(), "To element incorrect")

    def test_parse_recursive(self):
        text_xml = f"""
<section>
    <title>A simple section</title>
    <section>
        <title>One level section</title>
        <section xml:lang="esp">
            <title xml:lang="esp">Segundo nivel de recursión</title>
            <section>
                <title>Third level</title>
                <para>{self.PARAGRAPH}</para>
            </section>
        </section>
    </section>
</section>
        """
        section = EMLSection.from_string(text_xml)
        self.assertEqual("A simple section", section.title, "Title parsed incorrectly")
        self.assertEqual(Language.ENG, section.title.language, "Title language parsed incorrectly")
        self.assertEqual(Language.ENG, section.language, "Language parsed incorrectly")
        self.assertEqual("One level section", section.section.title, "Error on first recursion")
        self.assertEqual(Language.ENG, section.section.title.language, "Error on language on first recursion")
        self.assertEqual("Segundo nivel de recursión", section.section.section.title, "Error on second recursion")
        self.assertEqual(
            Language.ESP,
            section.section.section.title.language,
            "Error on language (title) on second recursion"
        )
        self.assertEqual(Language.ESP, section.section.section.language, "Error on language on second recursion")
        self.assertEqual("Third level", section.section.section.section.title, "Error on third recursion")
        self.assertEqual(
            Language.ENG,
            section.section.section.section.title.language,
            "Error on language on third recursion"
        )
        self.assertEqual(
            self.PARAGRAPH,
            section.section.section.section.paragraph,
            "Error on paragraph on third recursion"
        )


if __name__ == '__main__':
    unittest.main()



import datetime as dt
from typing import List, Union

from datetime_interval import Interval

from dwca.terms import Field


class LocationID(Field):
    """
    An identifier for the set of Location information.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/locationID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class HigherGeographyID(Field):
    """
    An identifier for the geographic region within which the Location occurred.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/higherGeographyID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class HigherGeography(Field):
    """
    A list of geographic names less specific than the information captured in the locality term.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/higherGeography"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Continent(Field):
    """
    The name of the continent in which the Location occurs.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/continent"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class WaterBody(Field):
    """
    The name of the water body in which the Location occurs.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/waterBody"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class IslandGroup(Field):
    """
    The name of the island group in which the Location occurs.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/islandGroup"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class Island(Field):
    """
    The name of the island on or near which the Location occurs.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/island"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class Country(Field):
    """
    The name of the country or major administrative unit in which the Location occurs.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/country"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class CountryCode(Field):
    """
    The standard code for the country in which the Location occurs.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/countryCode"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class StateProvince(Field):
    """
    The name of the next smaller administrative region than country in which the Location occurs.

    Smaller administrative region such as state, province, canton, department, region, etc.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/stateProvince"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class County(Field):
    """
    The full name of the next smaller administrative region than stateProvince in which the Location occurs.

    Such as state, province, canton, department, region, etc.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/county"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class Municipality(Field):
    """
    The full name of the next smaller administrative region than county in which the Location occurs.

    Such as city, municipality, etc. Do not use this term for a nearby named place that does not
    contain the actual Location.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/municipality"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCLocalityTerm(Field):
    """
    The specific description of the place.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/locality"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimLocality(Field):
    """
    The original textual description of the place.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimLocality"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MinimumElevationInMeters(Field):
    """
    The lower limit of the range of elevation (altitude, usually above sea level), in meters.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/minimumElevationInMeters"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MaximumElevationInMeters(Field):
    """
    The upper limit of the range of elevation (altitude, usually above sea level), in meters.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/maximumElevationInMeters"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimElevation(Field):
    """
    The original description of the elevation (altitude, usually above sea level) of the Location.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimElevation"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerticalDatum(Field):
    """
    The vertical datum used as the reference upon which the values in the elevation terms are based.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verticalDatum"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MinimumDepthInMeters(Field):
    """
    The lesser depth of a range of depth below the local surface, in meters.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/minimumDepthInMeters"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MaximumDepthInMeters(Field):
    """
    The greater depth of a range of depth below the local surface, in meters.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/maximumDepthInMeters"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimDepth(Field):
    """
    The original description of the depth below the local surface.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimDepth"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MinimumDistanceAboveSurfaceInMeters(Field):
    """
    The lesser distance in a range of distance from a reference surface in the vertical direction, in meters.

    Use positive values for locations above the surface, negative values for locations below.
    If depth measures are given, the reference surface is the location given by the depth,
    otherwise the reference surface is the location given by the elevation.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/minimumDistanceAboveSurfaceInMeters"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class MaximumDistanceAboveSurfaceInMeters(Field):
    """
    The greater distance in a range of distance from a reference surface in the vertical direction, in meters.

    Use positive values for locations above the surface, negative values for
    locations below. If depth measures are given, the reference surface is
    the location given by the depth, otherwise the reference surface is
    the location given by the elevation.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/maximumDistanceAboveSurfaceInMeters"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LocationAccordingTo(Field):
    """
    Information about the source of this Location information.

    Could be a publication (gazetteer), institution, or team of individuals.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/locationAccordingTo"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class LocationRemarks(Field):
    """
    Comments or notes about the Location.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/locationRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class DecimalLatitude(Field):
    """
    The geographic latitude of the geographic center of a Location.

    Latitude in decimal degrees, using the spatial reference system
    given in geodeticDatum. Positive values are north of the Equator,
    negative values are south of it. Legal values lie between -90
    and 90, inclusive.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/decimalLatitude"
    TYPE = float

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class DecimalLongitude(Field):
    """
    The geographic longitude of the geographic center of a Location.

    Longitude in decimal degrees, using the spatial reference system
    given in geodeticDatum. Positive values are east of the Greenwich
    Meridian, negative values are west of it. Legal values lie between
    -180 and 180, inclusive.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/decimalLongitude"
    TYPE = float

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class GeodeticDatum(Field):
    """
    The geodetic datum upon which the geographic coordinates given in decimalLatitude and decimalLongitude are based.

    Ellipsoid, geodetic datum, or spatial reference system (SRS).
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/geodeticDatum"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class CoordinateUncertaintyInMeters(Field):
    """
    The horizontal distance describing the smallest circle containing the whole of the Location.

    Horizontal distance in meters from the given latitude and
    longitude. Leave the value empty if the uncertainty is
    unknown, cannot be estimated, or is not applicable (because
    there are no coordinates). Zero is not a valid value for
    this term.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/coordinateUncertaintyInMeters"
    TYPE = float

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class CoordinatePrecision(Field):
    """
    A decimal representation of the precision of the coordinates given in the latitude and longitude.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/coordinatePrecision"
    TYPE = float

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class PointRadiusSpatialFit(Field):
    """
    The ratio of the area of the point-radius to the area of the spatial representation of the Location.

    Point-radius described by: {decimalLatitude, decimalLongitude, coordinateUncertaintyInMeters}
    Legal values are 0, greater than or equal to 1, or undefined. A value of
    1 is an exact match or 100% overlap. A value of 0 should be used if the
    given point-radius does not completely contain the original representation.
    The pointRadiusSpatialFit is undefined (and should be left empty) if the original
    representation is any geometry without area (e.g., a point or polyline) and
    without uncertainty and the given georeference is not that same geometry
    (without uncertainty). If both the original and the given georeference are the
    same point, the pointRadiusSpatialFit is 1.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/pointRadiusSpatialFit"
    TYPE = float

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimCoordinates(Field):
    """
    The verbatim original spatial coordinates of the Location.

    The coordinate ellipsoid, geodeticDatum, or full Spatial Reference
    System (SRS) for these coordinates should be stored in verbatimSRS
    and the coordinate system should be stored in verbatimCoordinateSystem.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimCoordinates"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimLatitude(Field):
    """
    The verbatim original latitude of the Location.

    The coordinate ellipsoid, geodeticDatum, or full Spatial Reference
    System (SRS) for these coordinates should be stored in verbatimSRS
    and the coordinate system should be stored in verbatimCoordinateSystem.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimLatitude"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimLongitude(Field):
    """
    The verbatim original longitude of the Location.

    The coordinate ellipsoid, geodeticDatum, or full Spatial Reference
    System (SRS) for these coordinates should be stored in verbatimSRS
    and the coordinate system should be stored in verbatimCoordinateSystem.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimLongitude"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimCoordinateSystem(Field):
    """
    The coordinate format for the verbatimLatitude and verbatimLongitude or the verbatimCoordinates of the Location.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimCoordinateSystem"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimSRS(Field):
    """
    The geodetic datum upon which coordinates given in (Verbatim) Latitude and Longitude, or Coordinates are based.

    The ellipsoid, geodetic datum, or spatial reference system (SRS)
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimSRS"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class FootprintWKT(Field):
    """
    A Well-Known Text (WKT) representation of the shape (footprint, geometry) that defines the Location.

    A Location may have both a point-radius representation (see :class:`DecimalLatitude`)
    and a footprint representation, and they may differ from each other.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/footprintWKT"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class FootprintSRS(Field):
    """
    The geodetic datum upon which the geometry given in footprintWKT is based.

    The ellipsoid, geodetic datum, or spatial reference system (SRS)
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/footprintSRS"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class FootprintSpatialFit(Field):
    """
    The ratio of the area of the footprintWKT to the area of the spatial representation of the Location.

    Legal values are 0, greater than or equal to 1, or undefined.
    A value of 1 is an exact match or 100% overlap. A value of 0
    should be used if the given footprintWKT does not completely
    contain the original representation. The footprintSpatialFit
    is undefined (and should be left empty) if the original
    representation is any geometry without area (e.g., a point or
    polyline) and without uncertainty and the given georeference is
    ot that same geometry (without uncertainty). If both the
    original and the given georeference are the same point,
    the footprintSpatialFit is 1.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/footprintSpatialFit"
    TYPE = float

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class GeoreferencedBy(Field):
    """
    A list of names of people, groups, or organizations who determined the georeference for the Location.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/georeferencedBy"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class GeoreferencedDate(Field):
    """
    The date on which the Location was georeferenced.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/georeferencedDate"
    TYPE = Union[dt.datetime, Interval]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class GeoreferenceProtocol(Field):
    """
    A description or reference to the methods used to determine the spatial footprint, coordinates, and uncertainties.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/georeferenceProtocol"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class GeoreferenceSources(Field):
    """
    A list of maps, gazetteers, or other resources used to georeference the Location.

    Described specifically enough to allow anyone in the
    future to use the same resources.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/georeferenceSources"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class GeoreferenceRemarks(Field):
    """
    Notes or comments about the spatial description determination.

    Explaining assumptions made in addition or opposition
    to the those formalized in the method referred to in
    georeferenceProtocol.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/georeferenceRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return
